import pickle
from collections import Counter
import re
import os

import pandas as pd
import numpy as np
from tqdm import tqdm, trange
import scipy.special


def norm_kmer_preds(kmer_preds):
    kmer_preds = np.log1p(kmer_preds)
    sums = kmer_preds.sum(axis=1, keepdims=True)
    norm_preds = kmer_preds.copy()
    norm_preds[sums[:, 0] != 0, :] /= sums[sums[:, 0] != 0, :]
    return norm_preds


def ensemble_predictions(kmer_preds_file, nn_preds):
    kmer_preds_df = pd.read_csv(kmer_preds_file)
    kmer_preds = kmer_preds_df.iloc[:, 1:].values

    nn_preds = np.stack(nn_preds, axis=0).mean(axis=0)
    assert kmer_preds.shape == nn_preds.shape

    best_kmer_weight = 0.1
    preds = norm_kmer_preds(np.log1p(kmer_preds)) * best_kmer_weight + \
            scipy.special.softmax(nn_preds, axis=1) * (1 - best_kmer_weight)

    sid_raw = []
    for _, row in kmer_preds_df.iterrows():
        sid_raw.append(re.sub(r"\.\d+$", "", row["sequence_id"]))

    res_df = kmer_preds_df.copy()
    res_df.iloc[:, 1:] = preds
    res_df["sequence_id"] = sid_raw
    res_df = res_df.set_index("sequence_id")

    sub_name = "outputs/submission.csv"
    res_df.to_csv(sub_name, index_label="sequence_id", float_format="%.5g")
