#pragma once

using namespace std;


class StdDevCalcKnuth {
    private:

        double m_count;
        double m_meanPrev, m_meanCurr;
        double m_sPrev, m_sCurr;
        double m_varianceCurr;

    public:
        StdDevCalcKnuth() {
            m_count = 0;
        }

        void append(double d) {
            // Calculate the variance based on equations (15) and (16) on
            // page 216 of "The Art of Computer Programming, Volume 2",
            // Second Edition. Donald E. Knuth.  Addison-Wesley
            // Publishing Company, 1973.
            //
            // The relationships between the variance, standard deviation,
            // and s are as follows
            //
            //                      s
            //     variance  =  -----------
            //                   count - 1
            //
            //                                -------------
            //                               /
            //     standard_deviation  =    /  variance
            //                            \/
            //
            m_count++;

            if (m_count == 1) {
                // Set the very first values.
                m_meanCurr     = d;
                m_sCurr        = 0;
                m_varianceCurr = m_sCurr;
            }
            else {
                // Save the previous values.
                m_meanPrev     = m_meanCurr;
                m_sPrev        = m_sCurr;

                // Update the current values.
                m_meanCurr     = m_meanPrev + (d - m_meanPrev) / m_count;
                m_sCurr        = m_sPrev    + (d - m_meanPrev) * (d - m_meanCurr);
                m_varianceCurr = m_sCurr / (m_count - 1);
            }
        }

    double get_std_dev() const {
        return sqrt(m_varianceCurr);
    }

    double get_mean() const {
    	return m_meanPrev;
    }
};


template <typename T>
vector<size_t> sort_indexes(const vector<T> &v) {

  // initialize original index locations
  vector<size_t> idx(v.size());
  iota(idx.begin(), idx.end(), 0);

  // sort indexes based on comparing values in v
  // using std::stable_sort instead of std::sort
  // to avoid unnecessary index re-orderings
  // when v contains elements of equal values
  stable_sort(idx.begin(), idx.end(),
       [&v](size_t i1, size_t i2) {return v[i1] < v[i2];});

  return idx;
}
