import pickle
from collections import Counter

import pandas as pd
import numpy as np
import torch
from tqdm import tqdm
from tqdm import tqdm


def convert_sequences(all_data):
    seqs = []
    alphabet = "ACGTN"
    for row in tqdm(all_data.itertuples(), total=len(all_data)):
        seqs.append([alphabet.index(letter) for letter in row.sequence])

    return seqs


def extract_bin_features(data, seqs):
    assert len(data) == len(seqs)
    bin_features = data.iloc[:, 2:].values.astype(np.float32)
    bin_features = np.concatenate((bin_features,
                                   np.log1p(np.array([len(s) for s in seqs], dtype=np.float32))[:, np.newaxis] * 0.1),
                                  axis=1)

    return bin_features


def load_features(features_file):
    print(f"Loading features from {features_file}")
    data = pd.read_csv(features_file)
    seqs = convert_sequences(data)
    seqs = [torch.LongTensor(s) for s in tqdm(seqs)]
    bin_features = extract_bin_features(data, seqs)
    return seqs, bin_features


def load_labels(labels_file):
    train_labels = pd.read_csv(labels_file)
    train_labels_matrix = train_labels.iloc[:, 1:].values
    train_labels = np.argmax(train_labels_matrix, axis=1)
    return train_labels
