# DrivenData: Genetic Engineering Attribution Challenge  2nd place solution

## Software requirements

* Python 3.7
* scikit-learn=0.22.1, tensorflow>=2.1, numpy, pandas (check requirements.txt)
* [BLAST](https://blast.ncbi.nlm.nih.gov/Blast.cgi?PAGE_TYPE=BlastDocs&DOC_TYPE=Download) installed

## Hardware requirements

* 128 GB of RAM
* NVIDIA GeForce RTX 2080 Ti 11 GB (or equivalent)
* 200 GB of space for training (BLAST predictions are large)

## Input files requirements

We expect input data located in '**../dataset/**' folder. We need the following files:
1) **train_values.csv**
2) **train_labels.csv**
3) **test_values.csv** - (important!) the same file as it was in competition we need this file for inference part because we used train + test matrix to create PCA features. Otherwise we will need to retrain models.
4) **holdout_values.csv** - file with new sequences you want to classify. Note: If you need to classify old test just copy it as 'holdout_values.csv'

## Inference part only

Extract content of **models.zip** in '**../models/**' directory. Put **holdout_values.csv** in '**../dataset/**' folder. 

#### Preprocessing

Our pipeline requires run of [BLAST](https://blast.ncbi.nlm.nih.gov/Blast.cgi?PAGE_TYPE=BlastDocs&DOC_TYPE=Download) program to match sequences like it was [in tutorial](https://altlabs.tech/geac/blast-benchmark/). 

* `python r01_holdout_create_blast_task.py [path to blast] [path to makedb]`

It will create task for blast as .bat shell files. [path to blast] and [path to makedb] optional parameters. Default values: _blastn_ and _makeblastdb_.

Then you need to run these created files:
* `sh ../cache/blast_task_holdout/step1_create_db.bat`
* `sh ../cache/blast_task_holdout/step2_run_0.bat`
* `sh ../cache/blast_task_holdout_vs_train_and_test/step1_create_db.bat`
* `sh ../cache/blast_task_holdout_vs_train_and_test/step2_run_0.bat`

After they finished you will have the following 2 files created:
1) ../cache/blast_res_holdout/result_fold_holdout_part_0.csv
2) ../cache/blast_res_holdout_vs_full/result_fold_holdout_vs_train_and_test_part_0.csv

* `python r02_holdout_convert_blast_preds_sorted_classes.py` - it will extract class match statistics for each holdout entry 
* `python r03_holdout_create_blast_matrices.py` - it will create set of BLAST-feature files extracted from blast files.
* `python r04_holdout_create_pca_from_blast_matrices.py` - it will reduce BLAST-feature files to 1024 components. You need PCA transformers located in **../models/pca/** folder on this step!

#### Inference step

After preprocessing you only need to run:

* `python inference.py --mode holdout`

Submission will be located in '**../predicts/submission_normalized.csv**' directory.

**Important note**: there will be created 2 files: submission.csv, submission_normalized.csv. Second file is normalized version of submission to have the same distribution of classes as train. 
It gives us slightly better score on LB so we used _normalized version as our best submission. If holdout is very small or its distribution is very like train distribution then _normalized version can be worse.

## Training part only

In this part you will create contents of **'../models/'** directory. After you have it you can switch to inference part.

#### Preprocessing

Out pipeline requires run of [BLAST](https://blast.ncbi.nlm.nih.gov/Blast.cgi?PAGE_TYPE=BlastDocs&DOC_TYPE=Download) program to match sequences like it was [in tutorial](https://altlabs.tech/geac/blast-benchmark/). 

* `python r11_train_create_blast_task.py [path to blast] [path to makedb]`

It will create task for blast as .bat shell files. [path to blast] and [path to makedb] optional parameters. Default values: _blastn_ and _makeblastdb_.

Then you need to run these created files:
* `sh ../cache/blast_task_full_vs_full_999/step1_create_db.bat`
* `sh ../cache/blast_task_full_vs_full_999/step2_run_0.bat`
* `sh ../cache/blast_task_full_vs_full_999/step2_run_1.bat`
* `sh ../cache/blast_task_full_vs_full_999/step2_run_2.bat`
* `sh ../cache/blast_task_full_vs_full_999/step2_run_3.bat`
* `sh ../cache/blast_task_full_vs_full_999/step2_run_4.bat`
* `sh ../cache/blast_task_full_vs_full_999/step2_run_5.bat`
* `sh ../cache/blast_task_full_vs_full_999/step2_run_6.bat`
* `sh ../cache/blast_task_full_vs_full_999/step2_run_7.bat`
* `sh ../cache/blast_task_full_vs_full_9999999/step1_create_db.bat`
* `sh ../cache/blast_task_full_vs_full_9999999/step2_run_0.bat`
* `sh ../cache/blast_task_full_vs_full_9999999/step2_run_1.bat`
* `sh ../cache/blast_task_full_vs_full_9999999/step2_run_2.bat`
* `sh ../cache/blast_task_full_vs_full_9999999/step2_run_3.bat`
* `sh ../cache/blast_task_full_vs_full_9999999/step2_run_4.bat`
* `sh ../cache/blast_task_full_vs_full_9999999/step2_run_5.bat`
* `sh ../cache/blast_task_full_vs_full_9999999/step2_run_6.bat`
* `sh ../cache/blast_task_full_vs_full_9999999/step2_run_7.bat`

Files `step2_run_*.bat` can (and expected to) be run in parallel. It's time consuming part. 

After they finished you will have the following 2 folders filled with 16 CSV files:
1) ../cache/blast_res_full_vs_full_999/*.csv
2) ../cache/blast_res_full_vs_full_9999999/*.csv

* `python r12_train_convert_blast_preds_sorted_classes.py` - it will extract class match statistics for each train entry
* `python r13_train_create_blast_matrices.py` - it will create set of BLAST-feature files extracted from blast files.
* `python r14_train_create_pca_from_blast_matrices.py` - it will reduce BLAST-feature files to 1024 components using PCA and create PCA transformers which will be used in inference.

#### Training step

After preprocessing you only need to run:

* `python training.py`

As result of training you will get set of *.h5 files in models directory. After training you can use Inference part 
which were described earlier, to get submission.csv with predictions.

**Note**: training can be done in parallel. You need to choose which configs to train in cycle. 
Otherwise all models will be trained one by one on single GPU.
