from data_providers.data_provider import DataProvider
from tensorflow.keras.models import load_model, Model
import pandas as pd
import importlib
import numpy as np
import os

def run(args, config_name):
    print('Args: {}'.format(args))
    model_config = importlib.import_module(f'configs.{config_name}').config

    try:
        model = load_model(args.models_path / f'{config_name}.h5', compile=False)
    except Exception as e:
        print('Exception 1: {}'.format(e))
        from training.architectures import create_arch_v1, create_arch_v2, create_arch_v3, create_arch_v4, create_arch_v5
        arch_version = model_config['arch_version']
        if arch_version == 1:
            create_arch = create_arch_v1
        elif arch_version == 2:
            create_arch = create_arch_v2
        elif arch_version == 3:
            create_arch = create_arch_v3
        elif arch_version == 4:
            create_arch = create_arch_v4
        elif arch_version == 5:
            create_arch = create_arch_v5
        model = create_arch(
            seq_kernels=model_config['seq_kernels'],
            seq_size=model_config['seq_size'],
            left_padding_size=10,
            embs_channels_size=model_config['embs_channels_size'],
        )
        model.load_weights(str(args.models_path / f'{config_name}.h5'))
        model = Model(inputs=model.inputs, outputs=model.layers[-2].output)


    data_provider = DataProvider(mode=args.mode,
                                 dataset_path=args.dataset_path,
                                 seq_size=model_config['seq_size'],
                                 left_padding_size=10)
    classes = data_provider.load_classes_names()
    predicts = []
    for nn_input in [data_provider.get_normal_seqs_input(), data_provider.get_reversed_seqs_input()]:
        predicts.append(model.predict(nn_input, batch_size=args.batch_size))
    # predicts = np.mean(predicts, axis=0)
    # my_submission = pd.DataFrame(data=predicts,
    #                              columns=classes,
    #                              index=data_provider.index)
    # my_submission.to_csv(args.predicts_path / f'{config_name}.csv')

    my_submission = pd.DataFrame(data=predicts[0],
                                 columns=classes,
                                 index=data_provider.index)
    if not os.path.isdir(str(args.predicts_path)):
        os.mkdir(str(args.predicts_path))
    my_submission.to_csv(args.predicts_path / f'{config_name}_{args.mode}.csv')

    my_submission = pd.DataFrame(data=predicts[1],
                                 columns=classes,
                                 index=data_provider.index)
    my_submission.to_csv(args.predicts_path / f'{config_name}_{args.mode}_reverse.csv')