from tensorflow.keras.layers import Embedding, Conv1D, Concatenate, GlobalMaxPooling1D, Dense, Input, \
    BatchNormalization, Dropout
from tensorflow.keras.models import Model
from tensorflow.keras import layers


def create_arch_v1(seq_kernels, left_padding_size, seq_size, embs_channels_size):
    processed_inputs = []

    input_alen = Input(1024, name='input_alen')
    x_alen = input_alen
    x_alen = BatchNormalization()(x_alen)
    x_alen = Dense(512, activation='relu')(x_alen)
    x_alen = BatchNormalization()(x_alen)
    x_alen = Dropout(0.5)(x_alen)
    x_alen = Dense(256, activation='relu')(x_alen)
    x_alen = BatchNormalization()(x_alen)
    x_alen = Dropout(0.5)(x_alen)

    input_blast = Input(1314, name='input_blast')
    x_blast = Dropout(0.25)(input_blast)
    x_blast = Dense(128, activation='relu')(x_blast)
    x_blast = BatchNormalization()(x_blast)
    x_blast = Dropout(0.25)(x_blast)

    input_cat = Input(39)
    x_cat = Dense(128, activation='relu')(input_cat)
    x_cat = BatchNormalization()(x_cat)
    x_cat = Dropout(0.1)(x_cat)
    x_cat = Dense(256, activation='relu')(x_cat)

    x_features = Concatenate(axis=1)([x_cat, x_blast, x_alen])
    x_features = BatchNormalization()(x_features)
    x_features = Dense(2048, activation='relu')(x_features)
    x_features = Dropout(0.25)(x_features)

    processed_inputs.append(x_features)
    input_seqs = Input(shape=(left_padding_size + seq_size,))
    embs = Embedding(8, 24, mask_zero=True)(input_seqs)
    for i in seq_kernels:
        conv = Conv1D(embs_channels_size, i, activation='elu')(embs)
        conv = GlobalMaxPooling1D()(conv)
        processed_inputs.append(conv)

    x = Concatenate(axis=1)(processed_inputs)
    x = Dense(1314)(x)
    x = layers.Softmax(dtype='float32')(x)
    return Model([input_seqs, input_cat, input_blast, input_alen], x)


def create_arch_v2(seq_kernels, left_padding_size, seq_size, embs_channels_size):
    seqs_embeddings = []

    input_seqs = Input(shape=(left_padding_size + seq_size,))
    embs = Embedding(8, 24, mask_zero=True)(input_seqs)
    for i in seq_kernels:
        conv = Conv1D(embs_channels_size, i, activation='elu')(embs)
        conv = GlobalMaxPooling1D()(conv)
        conv = Dropout(0.01)(conv)
        seqs_embeddings.append(conv)

    input_cat = Input(39, name='input_cat')
    x_cat = input_cat
    x_cat = Dense(128, activation='relu', name='cat1')(x_cat)
    x_cat = BatchNormalization()(x_cat)
    x_cat = Dropout(0.15)(x_cat)
    x_cat = Dense(256, activation='relu', name='cat2')(x_cat)
    x_cat = BatchNormalization()(x_cat)
    x_cat = Dropout(0.15)(x_cat)

    input_blast = Input(1314, name='input_blast')
    x_blast = input_blast
    x_blast = Dropout(0.25)(x_blast)
    x_blast = Dense(256, activation='relu')(x_blast)
    x_blast = BatchNormalization()(x_blast)
    x_blast = Dropout(0.25)(x_blast)

    input_gap = Input(1024, name='input_gap')
    x_gap = input_gap
    x_gap = Dense(512, activation='relu')(x_gap)
    x_gap = BatchNormalization()(x_gap)
    x_gap = Dropout(0.3)(x_gap)
    x_gap = Dense(128, activation='relu')(x_gap)
    x_gap = BatchNormalization()(x_gap)
    x_gap = Dropout(0.3)(x_gap)

    input_alen = Input(1024, name='input_alen')
    x_alen = input_alen
    x_alen = Dense(512, activation='relu')(x_alen)
    x_alen = BatchNormalization()(x_alen)
    x_alen = Dropout(0.3)(x_alen)
    x_alen = Dense(128, activation='relu')(x_alen)
    x_alen = BatchNormalization()(x_alen)
    x_alen = Dropout(0.3)(x_alen)

    x_pcas = Concatenate(axis=1)([x_alen, x_cat, x_blast, x_gap])
    x_pcas = Dense(1024 + 512, activation='relu')(x_pcas)
    x_pcas = Dropout(0.2)(x_pcas)

    seqs_embeddings.append(x_pcas)
    x = Concatenate(axis=1)(seqs_embeddings)

    x = Dense(1314, name='final_dense')(x)

    x = layers.Softmax(dtype='float32')(x)
    return Model([input_seqs, input_cat, input_blast, input_alen, input_gap], x)


def create_arch_v3(seq_kernels, left_padding_size, seq_size, embs_channels_size):
    seqs_embeddings = []

    input_seqs = Input(shape=(left_padding_size + seq_size,))
    embs = Embedding(8, 24, mask_zero=True)(input_seqs)
    for i in seq_kernels:
        conv = Conv1D(embs_channels_size, i, activation='elu')(embs)
        conv = GlobalMaxPooling1D()(conv)
        conv = Dropout(0.01)(conv)
        seqs_embeddings.append(conv)

    input_cat = Input(39, name='input_cat')
    x_cat = input_cat
    x_cat = Dense(128, activation='relu', name='cat1')(x_cat)
    x_cat = BatchNormalization()(x_cat)
    x_cat = Dropout(0.15)(x_cat)
    x_cat = Dense(256, activation='relu', name='cat2')(x_cat)
    x_cat = BatchNormalization()(x_cat)
    x_cat = Dropout(0.15)(x_cat)

    input_blast = Input(1314, name='input_blast')
    x_blast = input_blast
    x_blast = Dropout(0.4)(x_blast)
    x_blast = Dense(256, activation='relu')(x_blast)
    x_blast = BatchNormalization()(x_blast)
    x_blast = Dropout(0.1)(x_blast)

    input_gap = Input(1024, name='input_gap')
    x_gap = input_gap
    x_gap = Dense(512, activation='relu')(x_gap)
    x_gap = BatchNormalization()(x_gap)
    x_gap = Dropout(0.5)(x_gap)
    x_gap = Dense(128, activation='relu')(x_gap)
    x_gap = BatchNormalization()(x_gap)
    x_gap = Dropout(0.5)(x_gap)

    input_bitscore = Input(1024, name='input_bitscore')
    x_bitscore = input_bitscore
    x_bitscore = Dense(512, activation='relu')(x_bitscore)
    x_bitscore = BatchNormalization()(x_bitscore)
    x_bitscore = Dropout(0.5)(x_bitscore)
    x_bitscore = Dense(128, activation='relu')(x_bitscore)
    x_bitscore = BatchNormalization()(x_bitscore)
    x_bitscore = Dropout(0.5)(x_bitscore)

    input_mismatch = Input(1024, name='input_mismatch')
    x_mismatch = input_mismatch
    x_mismatch = Dense(512, activation='relu')(x_mismatch)
    x_mismatch = BatchNormalization()(x_mismatch)
    x_mismatch = Dropout(0.5)(x_mismatch)
    x_mismatch = Dense(128, activation='relu')(x_mismatch)
    x_mismatch = BatchNormalization()(x_mismatch)
    x_mismatch = Dropout(0.5)(x_mismatch)

    input_alen = Input(1024, name='input_alen')
    x_alen = input_alen
    x_alen = Dense(512, activation='relu')(x_alen)
    x_alen = BatchNormalization()(x_alen)
    x_alen = Dropout(0.5)(x_alen)
    x_alen = Dense(128, activation='relu')(x_alen)
    x_alen = BatchNormalization()(x_alen)
    x_alen = Dropout(0.5)(x_alen)

    x_pcas = Concatenate(axis=1)([x_alen, x_cat, x_blast, x_gap, x_bitscore, x_mismatch])
    x_pcas = Dense(2048, activation='relu')(x_pcas)
    x_pcas = Dropout(0.2)(x_pcas)

    seqs_embeddings.append(x_pcas)
    x = Concatenate(axis=1)(seqs_embeddings)

    x = Dense(1314, name='final_dense')(x)

    x = layers.Softmax(dtype='float32')(x)
    return Model([input_seqs, input_cat, input_blast, input_alen, input_gap, input_bitscore, input_mismatch], x)


def create_arch_v4(seq_kernels, left_padding_size, seq_size, embs_channels_size):
    seqs_embeddings = []

    input_seqs = Input(shape=(left_padding_size + seq_size,))
    embs = Embedding(8, 24, mask_zero=True)(input_seqs)
    for i in seq_kernels:
        conv = Conv1D(embs_channels_size, i, activation='elu')(embs)
        conv = GlobalMaxPooling1D()(conv)
        seqs_embeddings.append(conv)

    input_cat = Input(39, name='input_cat')
    x_cat = input_cat
    x_cat = Dense(128, activation='relu', name='cat1')(x_cat)
    x_cat = BatchNormalization()(x_cat)
    x_cat = Dense(256, activation='relu', name='cat2')(x_cat)
    x_cat = BatchNormalization()(x_cat)
    x_cat = Dense(512, activation='relu', name='cat3')(x_cat)
    x_cat = BatchNormalization()(x_cat)

    seqs_embeddings.append(x_cat)
    x = Concatenate(axis=1)(seqs_embeddings)
    x = Dense(1314, name='final_dense')(x)
    x = layers.Softmax(dtype='float32')(x)
    return Model([input_seqs, input_cat], x)


def create_arch_v5(seq_kernels, left_padding_size, seq_size, embs_channels_size):
    seqs_embeddings = []

    input_seqs = Input(shape=(left_padding_size + seq_size,))
    embs = Embedding(8, 24, mask_zero=True)(input_seqs)
    for i in seq_kernels:
        conv = Conv1D(embs_channels_size, i, activation='elu')(embs)
        conv = GlobalMaxPooling1D()(conv)
        seqs_embeddings.append(conv)
    if len(seqs_embeddings) > 1:
        x = Concatenate(axis=1)(seqs_embeddings)
    else:
        x = seqs_embeddings[0]
    x = Dense(1314, name='final_dense')(x)
    x = layers.Softmax(dtype='float32')(x)
    return Model([input_seqs], x)
