# -*- coding: utf-8 -*-
"""
Created on Sun Jan  3 22:09:36 2021

@author: IgorMehanika
"""

from kerastuner import HyperModel 
import tensorflow as tf
from tensorflow import keras

class CNNHyperModel(HyperModel):
    def __init__(self, input_shape, num_classes):
        self.input_shape = input_shape
        self.num_classes = num_classes

    def build(self, hp):
        model = keras.Sequential()
        model.add(
            tf.keras.layers.Conv2D(
                filters=hp.Choice(
                    'num_filters',
                    values=[9, 96],
                    default=32,
                ),
                kernel_size=1,
                strides = 1,
                activation='relu',
                input_shape=self.input_shape
            )
        )
        
        
        model.add(
            tf.keras.layers.Conv2D(
                filters=96,
                activation='relu',
                kernel_size=5,
                strides =1,
                padding='same'
            )
        )
        model.add(tf.keras.layers.BatchNormalization())
        model.add(tf.keras.layers.Activation('relu'))
        model.add(tf.keras.layers.MaxPooling2D(pool_size=(2, 2),
                                               strides=(2, 2)))        
        
        model.add(
            tf.keras.layers.Conv2D(
                filters=192,
                activation='relu',
                kernel_size=5,
                strides =1,
                padding='same'
            )
        )        
        
        model.add(tf.keras.layers.BatchNormalization())
        model.add(tf.keras.layers.Activation('relu'))
        model.add(tf.keras.layers.MaxPooling2D(pool_size=(2, 2),
                                               strides=(2, 2)))        
        
        
        model.add(
            tf.keras.layers.Conv2D(
                filters=384,
                activation='relu',
                kernel_size=5,
                strides =1,
                padding='same'
            )
        )        
        

        model.add(tf.keras.layers.BatchNormalization())
        model.add(tf.keras.layers.Activation('relu'))
        model.add(tf.keras.layers.MaxPooling2D(pool_size=(2, 2),
                                               strides=(2, 2)))
        
        model.add(
            tf.keras.layers.Conv2D(
                filters=576,
                activation='relu',
                kernel_size=5,
                strides =1,
                padding='same'
            )
        )        

        model.add(tf.keras.layers.BatchNormalization())
        model.add(tf.keras.layers.Activation('relu'))
        model.add(tf.keras.layers.MaxPooling2D(pool_size=(2, 2),
                                               strides=(2, 2)))
        
        model.add(
            tf.keras.layers.Conv2D(
                filters=768,
                activation='relu',
                kernel_size=5,
                strides =1,
                padding='same'
            )
        )

        
        model.add(tf.keras.layers.BatchNormalization())
        model.add(tf.keras.layers.Activation('relu'))
        model.add(tf.keras.layers.AveragePooling2D(pool_size=(8, 8), 
                                                   strides=(1, 1)))  
    
        model.add(tf.keras.layers.Flatten())
        
        model.add(
            tf.keras.layers.Dense(
                units=hp.Int(
                    'units',
                    min_value=768,
                    max_value=3115,
                    step=32,
                    default=768
                ),
                activation='tanh'
            )
        )
        
        model.add(tf.keras.layers.Dense(self.num_classes, activation='softmax'))       
    

        model.compile(
            optimizer=keras.optimizers.Adam(
                hp.Float(
                    'learning_rate',
                    min_value=0.0005,
                    max_value=0.0007,
                    sampling='LOG',
                    default=1e-3
                )
            ),
            loss='sparse_categorical_crossentropy',
            metrics=['accuracy']
        )
        return model

hypermodel = CNNHyperModel(input_shape=(128,128,1), num_classes=3115)




from kerastuner.tuners import RandomSearch


tuner = RandomSearch(
    hypermodel,
    objective='val_accuracy',
    seed=1,
    max_trials=10,
    executions_per_trial=2,
    directory='random_search',
    project_name='GenAtrib'
)


from kerastuner.tuners import Hyperband


tuner = Hyperband(
    hypermodel,
    max_epochs= 10,
    objective='val_accuracy',
    seed=1,
    executions_per_trial=2,
    directory='hyperband',
    project_name='GenAtrib'
)

tuner.search_space_summary()

N_EPOCH_SEARCH = 20

tuner.search(train_data, YTE, epochs=N_EPOCH_SEARCH,
             validation_data = (validation_data,YVE))

# Show a summary of the search
tuner.results_summary()

# Retrieve the best model.
BK1_model = tuner.get_best_models(num_models=1)[0]
# Evaluate the best model.
loss, accuracy = best_model.evaluate(validation_data,YVE)
best_hp1 = tuner.get_best_hyperparameters()[0]