# -*- coding: utf-8 -*-
"""
Created on Mon Jan  4 21:00:31 2021

@author: IgorMehanika
"""

import numpy as np
from sklearn import preprocessing
from skimage.transform import resize

##############################################################################
#SEQUENCE CONVERSION 
def let2num(nt):
    map =[1,11,2,12,0,0,3,13,0,0,7,0,8,15,0,0,0,5,9,4,4,14,10,15,6,0,16]
    nt = nt.lower()
    number =[ord(nt)+1-ord('a') for nt in nt]
    seq = [map[number-1] for number in number]
    return seq;
###############################################################################
def frameSeqin2frameInteg(data):
    preved = data
    for k in range(len(preved)):
        preved.sequence[k] = let2num(preved.sequence[k]) 
    return preved
##############################################################################
##############################################################################
#                       #WAVELET SECTION 
##############################################################################
##############################################################################
import inspect
import scipy.special
from scipy.fftpack import ifft
from scipy.fftpack import fft
from scipy.special import eval_genlaguerre
import warnings
###############################################################################
####### mod for rotate
def _rotate_main_angle(x):
    x = np.mod(x, 2 * np.pi)
    return np.where(x > np.pi, x - 2 * np.pi, x)
###############################################################################
def rotate(x):
    x = _rotate_main_angle(x)
    edge_cases = np.full_like(1j * x, np.nan)
    edge_cases[x == np.pi / 2] = 1j
    edge_cases[x == -np.pi / 2] = -1j
    edge_cases[np.logical_or(x == np.pi, x == -np.pi)] = -1
    edge_cases[np.logical_or(x == 0, np.logical_or(x == 2 * np.pi, x == -2 * np.pi))] = 1
    return np.where(np.isnan(edge_cases), np.exp(1j * x), edge_cases)
###############################################################################
#### time-domain wavelet to a frequency domain wavelet
def to_frequency_domain_wavelet(time_domain_wavelet, num_timepoints=None):
    if num_timepoints is not None:
        if time_domain_wavelet.shape[-1] < num_timepoints:
            padding = (num_timepoints - time_domain_wavelet.shape[-1]) // 2
            if padding * 2 + time_domain_wavelet.shape[-1] < num_timepoints:
                padding = (padding, padding + 1)
            else:
                padding = (padding, padding)
            pad_width = (0, 0) * (len(time_domain_wavelet.shape) - 1) + padding
            time_domain_wavelet = np.pad(time_domain_wavelet, pad_width, mode='constant')
        elif time_domain_wavelet.shape[-1] > num_timepoints:
            time_domain_wavelet = time_domain_wavelet[..., :num_timepoints]
            raise ValueError('unexpected') 
######ray USING fft OK
    psi_f = fft(time_domain_wavelet)
    omega = 2 * np.pi * np.linspace(0, 1 - (1 / psi_f.shape[-1]), psi_f.shape[-1])
    omega = np.reshape(omega, (1,) * (len(psi_f.shape) - 1) + (omega.shape[0],))
    if psi_f.shape[-1] // 2 * 2 == psi_f.shape[-1]:  # is even
        psi_f = psi_f * rotate(-omega * (psi_f.shape[-1] + 1) / 2) * np.sign(np.pi - omega)
    else:
        psi_f = psi_f * rotate(-omega * (psi_f.shape[-1] + 1) / 2)
    return psi_f
###############################################################################
def _awt(x, psi_f, is_time_domain_wavelet_real):
    # x = (batch, time)
    # psi_f = (..., scale, time)
    if not np.isrealobj(x):
        x = x / np.sqrt(2)
    psi_f = np.conj(psi_f)
    result = np.expand_dims(fft(np.where(np.isnan(x), 0, x)), 1) * np.reshape(psi_f, (1, -1, psi_f.shape[-1]))
    result = np.reshape(result, (x.shape[0],) + psi_f.shape)
    x = np.reshape(x, (x.shape[0],) + (1,) * (len(result.shape) - 2) + (x.shape[1],))
    result = ifft(result)
    if np.isrealobj(x) and is_time_domain_wavelet_real and not np.isrealobj(result):
        result = np.real(result)
    if not np.any(np.isfinite(result)):
        if not np.isrealobj(result):
            result = np.inf * (1 + 1j) * np.ones_like(result)
        else:
            result = np.inf * np.ones_like(result)
    if not np.isrealobj(result):
        result = np.where(np.isnan(x), np.nan * (1 + 1j), result)
    else:
        result = np.where(np.isnan(x), np.nan, result)
    return result
###############################################################################
# using freq. wav
def analytic_wavelet_transform(x, frequency_domain_wavelet, is_time_domain_wavelet_real, unpad_slices=None):
    x = np.asarray(x)
    frequency_domain_wavelet = np.asarray(frequency_domain_wavelet)

    if x.shape[-1] != frequency_domain_wavelet.shape[-1]:
        raise ValueError(
            'x and frequency_domain_wavelet must have same shape.'
            'x.shape: {}, frequency_domain_wavelet.shape: {}'.format(x.shape, frequency_domain_wavelet.shape))

    if unpad_slices is not None:
        if not isinstance(unpad_slices, (tuple, list)):
            unpad_slices = (unpad_slices,)
        else:
            unpad_slices = tuple(unpad_slices)
        if len(unpad_slices) != x.ndim:
            raise ValueError('DFimension mismatch. For {} unpad_slices, and x has {} axes'.format(
                len(unpad_slices), x.ndim))
    x_shape = x.shape
    x = np.reshape(x, (-1, x.shape[-1]))

    indicator_good = np.sum(np.isfinite(x), axis=-1) > 1
    if np.sum(indicator_good) == 0:
        # make a dummy result of the correct size
        result = _awt(np.zeros((1, x.shape[-1]), dtype=x.dtype), frequency_domain_wavelet, is_time_domain_wavelet_real)
        result = np.inf * (1 + 1j) * np.tile(result, (x.shape[0],) + (1,) * result.shape[1:])
    else:
        x = x[indicator_good]
        result_good = _awt(x, frequency_domain_wavelet, is_time_domain_wavelet_real)
        result = np.full((x.shape[0],) + result_good.shape[1:], np.inf, dtype=result_good.dtype)
        result[indicator_good] = result_good
    result = np.reshape(result, x_shape[:-1] + result.shape[1:])
    if unpad_slices is not None:
        s = list(unpad_slices[:-1])
        for _ in range(len(unpad_slices), result.ndim):
            s.append(slice(None))
        s.append(unpad_slices[-1])
        s = tuple(s)
        assert(len(s) == result.ndim)
        result = result[s]
    return result
###############################################################################
#analitical options
def analytic_transform(x, is_output_frequency=False):

    z = fft(np.where(np.isnan(x), 0, x))
    if np.isrealobj(x):
        z = 2 * z
    if x.shape[-1] // 2 * 2 == x.shape[-1]:  # even
        indices = np.arange(x.shape[-1] // 2 + 1, z.shape[-1])
        np.put_along_axis(z, np.reshape(indices, (1,) * (len(z.shape) - 1) + (-1,)), 0, -1)
        np.put_along_axis(
            z,
            np.reshape(np.array([x.shape[-1] - 1]), (1,) * (len(z.shape) - 1) + (-1,)),
            np.take(z, x.shape[-1] // 2, -1),
            -1)
    else:
        indices = np.arange((x.shape[-1] + 1) // 2, z.shape[-1])
        np.put_along_axis(z, np.reshape(indices, (1,) * (len(z.shape) - 1) + (-1,)), 0, -1)
    if not is_output_frequency:
        z = ifft(z)
        if z.shape[-1] != x.shape[-1]:
            z = np.take(z, np.arange(0, x.shape[-1]), -1)
    return z
##############################################################################
##############################################################################
# OBJECT AMORS USING G.MORSE G.MORSE<AMORS OR AMORS(G.MORSE)
class GeneralizedMorseWavelet:
##############################################################################
    def __init__(self, gamma, beta, is_bandpass_normalized=True):
        #  make gamma and beta same size
        self._gamma = np.zeros_like(beta) + gamma
        self._beta = np.zeros_like(gamma) + beta
        self._is_bandpass_normalized = is_bandpass_normalized
##############################################################################
#METHODS HERE
    @staticmethod
    def replace(instance, **kwargs):
        property_names = [n for n, v in inspect.getmembers(type(instance), lambda m: isinstance(m, property))]
        init_kwargs = inspect.getfullargspec(type(instance).__init__).args
        replaced = dict()
        for k in init_kwargs[1:]:
            if k in kwargs:
                replaced[k] = kwargs[k]
            elif k in property_names:
                replaced[k] = getattr(instance, k)
        return type(instance)(**replaced)
##############################################################################
#PROPERTIES HERE
    @property
    def gamma(self):
        return self._gamma

    @property
    def beta(self):
        return self._beta

    @property
    def is_bandpass_normalized(self):
        return self._is_bandpass_normalized
##############################################################################
    def time_domain_width(self):
        # W = pi * number of oscillations in the the range of central wav.window
        return np.sqrt(self.beta * self.gamma)
##############################################################################
    def footprint(self, scale_frequency, standard_widths=4):

        # N = number of points holding ~95% of wavelet energy.
        # aproximate 2 * sqrt(2) * time_domain_width / scale_frequency close to needed ~95%
        # 2 * sqrt(2) * time_domain_width / scale_frequency close to needed ~95%
        footprint = np.sqrt(2) * standard_widths / 2 * self.time_domain_width() / scale_frequency
        return np.ceil(footprint).astype(int)
##############################################################################
    def demodulated_skewness_imag(self):
        return (self.gamma - 3) / self.time_domain_width()
##############################################################################
    def demodulated_kurtosis(self):
        return 3 - np.power(self.demodulated_skewness_imag(), 2) - (2 / np.power(self.time_domain_width(), 2))
##############################################################################
    def peak_frequency(self):
        with warnings.catch_warnings():
            warnings.filterwarnings('ignore', category=RuntimeWarning)
            return np.where(
                self.beta == 0,
                np.power(np.log(2), 1 / self.gamma),
                np.exp((1 / self.gamma) * (np.log(self.beta) - np.log(self.gamma))))
##############################################################################
    def energy_frequency(self):
        return (1 / np.power(2, 1 / self.gamma)) * (
                scipy.special.gamma((2 * self.beta + 2) / self.gamma)
                / scipy.special.gamma((2 * self.beta + 1) / self.gamma))
##############################################################################
    def instantaneous_frequency(self):
        return scipy.special.gamma((self.beta + 2) / self.gamma) / scipy.special.gamma((self.beta + 1) / self.gamma)
##############################################################################
    def curvature_instantaneous_frequency(self):
        _, frequency_cumulants = self.frequency_domain_cumulants(3)
        return frequency_cumulants[3] / np.sqrt(np.power(frequency_cumulants[2], 3))
##############################################################################
    def heisenberg_box(self):
        beta = np.where(self.beta == 1, 1 + 1e-10, self.beta)
        beta = np.where(beta < 0.5, np.nan, beta)

        with warnings.catch_warnings():
            warnings.filterwarnings('ignore', category=RuntimeWarning)
            log_sigma_freq_1 = ((2 / self.gamma) * np.log(self.gamma / (2 * beta))
                                + scipy.special.gammaln((2 * beta + 1 + 2) / self.gamma)
                                - scipy.special.gammaln((2 * beta + 1) / self.gamma))
            log_sigma_freq_2 = ((2 / self.gamma) * np.log(self.gamma / (2 * beta))
                                + 2 * scipy.special.gammaln((2 * beta + 2) / self.gamma)
                                - 2 * scipy.special.gammaln((2 * beta + 1) / self.gamma))

            sigma_freq = np.sqrt(np.exp(log_sigma_freq_1) - np.exp(log_sigma_freq_2))

            def _log_a(g, b):
                return (b / g) * (1 + np.log(g) - np.log(b))

            ra = (2 * _log_a(self.gamma, beta)
                  - 2 * _log_a(self.gamma, beta - 1)
                  + _log_a(self.gamma, 2 * (beta - 1))
                  - _log_a(self.gamma, 2 * beta))
            rb = (2 * _log_a(self.gamma, beta)
                  - 2 * _log_a(self.gamma, beta - 1 + self.gamma)
                  + _log_a(self.gamma, 2 * (beta - 1 + self.gamma))
                  - _log_a(self.gamma, 2 * beta))
            rc = (2 * _log_a(self.gamma, beta)
                  - 2 * _log_a(self.gamma, beta - 1 + self.gamma / 2)
                  + _log_a(self.gamma, 2 * (beta - 1 + self.gamma / 2))
                  - _log_a(self.gamma, 2 * beta))

            log_sigma_2a = (ra
                            + (2 / self.gamma) * np.log((beta / self.gamma))
                            + 2 * np.log(beta)
                            + scipy.special.gammaln((2 * (beta - 1) + 1) / self.gamma)
                            - scipy.special.gammaln((2 * beta + 1) / self.gamma))
            log_sigma_2b = (rb
                            + (2 / self.gamma) * np.log((beta / self.gamma))
                            + 2 * np.log(self.gamma)
                            + scipy.special.gammaln((2 * (beta - 1 + self.gamma) + 1) / self.gamma)
                            - scipy.special.gammaln((2 * beta + 1) / self.gamma))
            log_sigma_2c = (rc
                            + (2 / self.gamma) * np.log(beta / self.gamma)
                            + np.log(2) + np.log(beta) + np.log(self.gamma)
                            + scipy.special.gammaln((2 * (beta - 1 + self.gamma / 2) + 1) / self.gamma)
                            - scipy.special.gammaln((2 * beta + 1) / self.gamma))

            sigma_time = np.sqrt(np.exp(log_sigma_2a) + np.exp(log_sigma_2b) - np.exp(log_sigma_2c))
            sigma_time = np.where(np.isnan(beta), np.nan, sigma_time)
            sigma_time = np.real(sigma_time)
            sigma_freq = np.real(sigma_freq)

            area = sigma_time * sigma_freq

        return area, sigma_time, sigma_freq
##############################################################################
# logpoints or
    def log_spaced_frequencies(
            self,
            num_timepoints=None,
            high=None,
            low=None,
            nyquist_overlap=None,
            endpoint_overlap=None,
            density=4):
 # OPTION 

        with warnings.catch_warnings():
            warnings.filterwarnings('ignore', category=RuntimeWarning)
            if high is None and nyquist_overlap is None:
                high = np.pi
                nyquist_overlap = 0.1
            if low is None and endpoint_overlap is None:
                if num_timepoints is None:
                    raise ValueError('num_timepoints must be specified.')
                endpoint_overlap = 5
            if endpoint_overlap is not None and num_timepoints is None:
                raise ValueError('num_timepoints must be specified.')

            r = 1 + (1 / (density * self.time_domain_width()))

            if nyquist_overlap is not None:
                nyquist_high = self._high_frequency_cutoff(nyquist_overlap)
                if high is None:
                    high = nyquist_high
                else:
                    high = np.where(nyquist_high < high, nyquist_high, high)
            if endpoint_overlap is not None:
                endpoint_low = self._low_frequency_cutoff(endpoint_overlap, num_timepoints)
                if low is None:
                    low = endpoint_low
                else:
                    low = np.where(endpoint_low > low, endpoint_low, low)

            n = np.floor(np.log(high / low) / np.log(r)).astype(int)

            if not np.ndim(n) == 0:
                indices = np.reshape(np.arange(np.max(n) + 1), (1,) * len(n.shape) + (np.max(n) + 1,))
                indices = np.where(indices < np.expand_dims(n, -1), indices, np.nan)
                return np.expand_dims(high, 1) / np.power(r, indices)
            return high / np.power(r, np.arange(n + 1))
##############################################################################
    def _high_frequency_cutoff(self, eta):
        # helper for log_spaced_frequencies
        omega_high = np.reshape(np.linspace(0, np.pi, 10000), (1,) * len(self.gamma.shape) + (-1,))
        peak_frequency = self.peak_frequency()
        if not np.ndim(peak_frequency) == 0:
            peak_frequency = np.expand_dims(peak_frequency, -1)
        # self.gamma.shape + (10000,)
        omega = peak_frequency * np.pi / omega_high
        ln_psi_1 = (self.beta / self.gamma) * np.log((np.exp(1) * self.gamma) / self.beta)
        beta, gamma = self.beta, self.gamma
        if not np.ndim(self.beta) == 0:
            beta, gamma = np.expand_dims(self.beta, -1), np.expand_dims(self.gamma, -1)
        ln_psi_2 = beta * np.log(omega) - np.power(omega, gamma)
        ln_psi = ln_psi_1 + ln_psi_2
        indices = np.tile(
            np.reshape(np.arange(ln_psi.shape[-1]), (1,) * (len(ln_psi.shape) - 1) + (ln_psi.shape[-1],)),
            ln_psi.shape[:-1] + (1,))
        indices = np.where(np.log(eta) - ln_psi <= 0, indices, np.nan)
        indices = np.nanmin(indices, axis=-1).astype(int)
        f = np.reshape(omega_high[np.reshape(indices, -1)], indices.shape)
        if np.ndim(gamma) == 0:
            return f.item()
        return f
##############################################################################
    def _low_frequency_cutoff(self, r, num_timepoints):
        # helper for log_spaced_frequencies
        return (2 * np.sqrt(2) * self.time_domain_width() * r) / num_timepoints
##############################################################################
    def amplitude(self, orthogonal_family_order=1):
        if self.is_bandpass_normalized:
            if not np.ndim(self.beta) == 0:
                om = self.peak_frequency()
                with warnings.catch_warnings():
                    warnings.filterwarnings('ignore', category=RuntimeWarning)
                    result = 2 / np.exp(self.beta * np.log(om) - np.power(om, self.gamma))
                result[self.beta == 0] = 2
            elif self.beta == 0:
                result = 2
            else:
                om = self.peak_frequency()
                result = 2 / np.exp(self.beta * np.log(om) - np.power(om, self.gamma))
        else:
            r = (2 * self.beta + 1) / self.gamma
            result = np.power(
                2 * np.pi * self.gamma * np.power(2, r)
                * np.exp(
                    scipy.special.gammaln(orthogonal_family_order)
                    - scipy.special.gammaln(orthogonal_family_order + r - 1)),
                1 / 2)
        return result
##############################################################################
    @staticmethod
    def _moments_to_cumulants(moments):
        result = np.full(moments.shape, np.nan)
        result[0] = np.log(moments[0])
        for idx in range(1, len(moments)):
            coefficients = np.zeros_like(result[0])
            for k in range(1, idx):
                coefficients += scipy.special.binom(idx - 1, k - 1) * result[k] * (moments[idx - k] / moments[0])
            result[idx] = (moments[idx] / moments[0]) - coefficients
        return result
##############################################################################
    def frequency_domain_moment(self, order):
        return (self.amplitude()
                * (1 / (2 * np.pi * self.gamma) * scipy.special.gamma((self.beta + order + 1) / self.gamma)))
##############################################################################
    def energy_moment(self, order):
        return ((2 / np.power(2, (order + 1) / self.gamma))
                * GeneralizedMorseWavelet.replace(self, beta=2*self.beta).frequency_domain_moment(order))
##############################################################################
#SANETY comare TOOLOOPY
    def frequency_domain_cumulants(self, max_order):
        moments = None
        for order in range(max_order + 1):
            moment = self.frequency_domain_moment(order)
            if moments is None:
                moments = np.full((max_order + 1,) + moment.shape, np.nan)
            moments[order] = moment
        return moments, GeneralizedMorseWavelet._moments_to_cumulants(moments)
##############################################################################
    def energy_cumulants(self, max_order):
        moments = None
        for order in range(max_order + 1):
            moment = self.energy_moment(order)
            if moments is None:
                moments = np.full((max_order + 1,) + moment.shape, np.nan)
            moments[order] = moment
        return moments, GeneralizedMorseWavelet._moments_to_cumulants(moments)
##############################################################################
#WAVELET FACTORY
    def make_wavelet(self, num_timepoints, scale_frequencies, num_orthogonal_family_members=1):

        scale_frequencies = np.asarray(scale_frequencies)

        if len(scale_frequencies.shape) > 1:
            raise ValueError('scales must be 1d vector')

        if np.ndim(scale_frequencies) == 0:
            scale_frequencies = np.expand_dims(scale_frequencies, 0)

        fo = np.reshape(self.peak_frequency(), self.gamma.shape + (1, 1))
        gamma = np.reshape(self.gamma, self.gamma.shape + (1, 1))
        beta = np.reshape(self.beta, self.beta.shape + (1, 1))

        # ([1, ..., 1,] freq, 1)
        scale_frequencies_reshaped = np.reshape(scale_frequencies, (1,) * (len(fo.shape) - 2) + (-1, 1))
        # fo.shape + (freq, 1)
        fact = np.abs(scale_frequencies_reshaped) / fo
        omega = 2 * np.pi * np.linspace(0, 1 - (1 / num_timepoints), num_timepoints)
        # fo.shape + (freq, time)
        omega = np.reshape(omega, (1,) * (len(fact.shape) - 1) + (-1,)) / fact

        with warnings.catch_warnings():
            warnings.filterwarnings('ignore', category=RuntimeWarning)
            if self.is_bandpass_normalized:
                psi_zero = np.where(
                    beta == 0,
                    2 * np.exp(-omega ** gamma),
                    2 * np.exp(-beta * np.log(fo) + fo ** gamma + beta * np.log(omega) - omega ** gamma))
            else:
                psi_zero = np.where(
                    beta == 0,
                    np.exp(-omega ** gamma),
                    np.exp(beta * np.log(omega) - omega ** gamma))

        # psi_zero shape is fo.shape + (freq, time)
        psi_zero[..., 0] = 1 / 2 * psi_zero[..., 0]
        psi_zero = np.where(np.isnan(psi_zero), 0, psi_zero)

        psi_f = self._first_family(fact, num_timepoints, num_orthogonal_family_members, omega, psi_zero)
        psi_f = np.where(np.isinf(psi_f), 0, psi_f)
        if len(psi_f.shape) > len(omega.shape):
            # add the family-order axis
            omega = np.expand_dims(omega, -3)
            fact = np.expand_dims(fact, -3)

        psi = ifft(psi_f * rotate(omega * (num_timepoints + 1) / 2 * fact))

        indicator_neg_sf = scale_frequencies < 0

        psi[..., indicator_neg_sf, :] = np.conj(psi[..., indicator_neg_sf, :])
        psi_f[..., indicator_neg_sf, 1:] = np.flip(psi_f[..., indicator_neg_sf, 1:], axis=-1)

        return psi, psi_f
##############################################################################
# FOR FINDING BETER LOCALIZATORS    
    def _first_family(self, fact, num_timepoints, num_orthogonal_family_members, omega, psi_zero):
        beta = np.reshape(self.beta, self.beta.shape + (1, 1))
        gamma = np.reshape(self.gamma, self.gamma.shape + (1, 1))
        r = (2 * beta + 1) / gamma
        c = r - 1

        # ([1, ..., 1,] freq, time)
        laguerre = np.zeros_like(omega)
        indices = np.arange(int(np.ceil(num_timepoints / 2)))
        # (wavelet, ..., freq, time)
        psi_f = np.zeros(((num_orthogonal_family_members,) + psi_zero.shape), dtype=psi_zero.dtype)

        for k in range(num_orthogonal_family_members):
            if self.is_bandpass_normalized:
                coeff = np.where(
                    beta == 0,
                    1,
                    np.sqrt(
                        np.exp(scipy.special.gammaln(r) + scipy.special.gammaln(k + 1) - scipy.special.gammaln(k + r))))
            else:
                # (freq, time)
                coeff = np.sqrt(1 / fact) * self.amplitude(k + 1)
            laguerre[..., indices] = eval_genlaguerre(k, c, 2 * np.power(omega[..., indices], gamma))
            psi_f[k] = coeff * psi_zero * laguerre
        # either eliminate the family-order axis or move the family-order axis to -3
        if num_orthogonal_family_members == 1:
            psi_f = np.squeeze(psi_f, 0)
        else:
            psi_f = np.moveaxis(psi_f, 0, -3)
        # (..., [wavelet,], freq, time)
        return psi_f
# ###############################################################################
# ###############################################################################
# # DEFINE NEEDED WAVELET / constants for amor wavelet acording to sequnce and voc number 32
# ###############################################################################
# ###############################################################################
# def calcamorse(presek,voc):
#     #wavelet functions constants
#     a =  -1.946115668748273e+05
#     b =  -2.376496184989404e-04
#     c =   1.944934189837382e+05
    
#     p1 = -0.0622058729465507
#     p2 = 2.657619574063227e+02
#     p3 = -1.479054694938970e+05
#     p4 = 23.6253072239594e+006
#     p5 = -3.529610946929338e+07
#     q1 = 1.595864246623817e+02
#     q2 = 2.425752421282320e+04
#     q3 = 1.547889826205923e+06
#     q4 = -2.390715581077701e+06
#     # scaling number
#     scnumb = 1 + round(a*len(presek)**b+c)
#     x = np.array(range(1,scnumb), dtype=float)
#     frs = np.true_divide((p1*(x**4) + p2*(x**3) + p3*(x**2) + p4*x + p5),((x**4) + q1*(x**3) + q2*(x**2) + q3*x + q4))
#     frselekt = 2*np.pi* frs/voc;#SHIFTING 
#     return frselekt
# #############################################################################
# # TEST TEST TEST run on single example with position red
# # red=1
# # presek = pre.sequence[red]
# # frselekt = 2*np.pi/np.logspace(np.log10(10),np.log10(100),292)
# # frselekt = calcamorse(pre.sequence[red],32)
# ###############################################################################
# #wav scales
# def calcscales(presek,voc):
#     a =  -1.946115668748273e+05
#     b =  -2.376496184989404e-04
#     c =    1.944934189837382e+05

#     factor = 2**(1/voc)

#     genscales = np.array(range(0,600),dtype=float)
#     genscales[0] =  0.689988916500468

#     for kk in range(1,600):
#         genscales[kk]=factor*genscales[kk-1]

#     scnumb = round(a*len(presek)**b+c)
#     scales = genscales[range(scnumb)]
#     return scales
# ###############################################################################
# #### MAKE SCALOGRAMS (scalograms are shifted, see coment SHIFTED on line 585)
# # Scalograms are produced using custom functions defined above that calculate 
# # tranformation coeficients W.
# # For this challenge a so-called analytical wavelet is used to achieve
# # localization abilities, several such wavelets are known,
# # here only "armor" (analytical morse) wavelet is defined and used
# # with parameters ensuring good localization properties.
# # A slight shift to scalogram image is introduced for the PyTorch implementation
# # (network exchange between platforms is desired as part of open code practice,
# # and you might desire to quickly retrain the model using PyTorch).
# # Since uses one side padding practices that are not that so friendly for scalogram, 
# # vertical shifting is introduced. TO REMOVE THIS USE VALUE FORMED ON LINE 584 (frs)
# # and not frselect on 585 (SHIFTED VERSION).
# # Shift won't affect other implementations like TF Keras, Matlab, SAS…. 
# # This shift in the vertical direction is achieved by scaling the scaling parameters 
# # (moving focus more toward details sort of a higher resolution
# #  + zoom effect trick for feature extraction and composing using wavelets). 
# # PLEASE note that the network average pooling layer in model to be used introduces insensitivity
# # to the position of features that reveal the lab origin,
# # but this “moving the image inside the frame” a bit further from the edge (seen as a jumping scalogram on images)
# # keeps things robust against potentially confusing effects for the learning process
# # caused by such padding (and for everything else, business as usual).
# ##############################################################################
# #PRODUCING SCALOGRAM FROM COEFS.
# from skimage.transform import resize
# def scalogramarea(w):
 
#     coef=resize(abs(w),(128,128), mode='constant')    
#     skal = preprocessing.MinMaxScaler(feature_range=(0,1))
#     skal=skal.fit(coef)
#     norm = skal.transform(coef)

#     im = np.uint8(norm*255)
#     return im

#FOR THIS CHALANGE SPECIFIC SHIFTED SCALOGRAM
#wavelet definition
def calcamorse(presek,voc):
    a =  -1.946115668748273e+05
    b =  -2.376496184989404e-04
    c =   1.944934189837382e+05
    p1 = -0.0622058729465507
    p2 = 2.657619574063227e+02
    p3 = -1.479054694938970e+05
    p4 = 23.6253072239594e+006
    p5 = -3.529610946929338e+07
    q1 = 1.595864246623817e+02
    q2 = 2.425752421282320e+04
    q3 = 1.547889826205923e+06
    q4 = -2.390715581077701e+06
    scnumb = 1 + round(a*len(presek)**b+c)
    x = np.array(range(1,scnumb), dtype=float)
    frs = np.true_divide((p1*(x**4) + p2*(x**3) + p3*(x**2) + p4*x + p5),((x**4) + q1*(x**3) + q2*(x**2) + q3*x + q4))
    frselekt = 2*np.pi*frs/voc;
    return frselekt
##############################################################################
#scal.wave calculation
def calcscales(presek,voc):
    a =  -1.946115668748273e+05
    b =  -2.376496184989404e-04
    c =    1.944934189837382e+05

    factor = 2**(1/voc)
    genscales = np.array(range(0,600),dtype=float)
    genscales[0] =  0.689988916500468/(2*np.pi)
    #genscales[0] =  2.592941518685188
    for kk in range(1,600):
        genscales[kk]=factor*genscales[kk-1]
    scnumb = round(a*len(presek)**b+c)
    scales = genscales[range(scnumb)]
    return scales
##############################################################################
#tranform coefs
def scalogramarea(w):
    coef = resize(abs(w),(128,128), mode='constant')    
    skal = preprocessing.MinMaxScaler(feature_range=(0,1))
    skal = skal.fit(coef)
    norm = skal.transform(coef)
    im = np.uint8(norm*255)
    return im
##############################################################################
def iter(sequences,red):
    Xsc = np.ndarray(shape = (red,128,128))#alocation
    for red in range(red):
        scales = calcscales(sequences[red],32)
        psi,psi_f= GeneralizedMorseWavelet(3, 120).make_wavelet(len(sequences[red]), scales)
        wx = analytic_wavelet_transform(sequences[red],psi_f, np.isrealobj(psi))
        Xsc[red,:,:] = scalogramarea(wx)
    return Xsc