﻿using System;
using System.Linq;
using GenEngAttr.Features;
using GenEngAttr.Models;
using GenEngAttr.Utils;

namespace GenEngAttr.Data
{
    // read/write of dataset and transforms
    public class DataSet
    {
        public static string GenAttrDir = @"c:\GenEngAttr\data"; 

        public static int NTest { get { return (NTestValue >= 0) ? NTestValue : ReadNTest(); } } //NTest = 18816; 

        public const int TopN = 10;
        public const int NLabs = 1314;
        public const int NTrain = 63017; 
        public const int NColumns = 39;
        public const int NColumnsAll = NColumns + 7;
        public const int NThreads6 = 6;

        private static int NTestValue = -1;

        public static string GenAttrSubmitDir { get { return GenAttrDir + @"submits\"; } }

        public static string GenAttrCacheDir { get { return GenAttrDir + @"cache\"; } } 
        public static string GenAttrInputsDir { get { return GenAttrCacheDir + @"inputs\"; } }
        public static string GenAttrFeaturesDir { get { return GenAttrCacheDir + @"features\"; } }
        public static string GenAttrOutModelsDir { get { return GenAttrCacheDir + @"outmodels\"; } }
        public static string GenAttrTrain { get { return GenAttrDir + "train_values.csv"; } }
        public static string GenAttrTest { get { return GenAttrDir + "test_values.csv"; } }
        public static string GenAttrLabel { get { return GenAttrDir + "train_labels.csv"; } } 

        public static readonly ulong[] GroupsMasks = { 0x1FUL, 0xE0UL, 0x7FF00UL, 0x1FF80000UL, 0x7FE0000000UL }; // 0x7FFFFFFFFFUL

        public static bool SetupDataDirectory(string[] args)
        {
            if ((args.Length > 0) && (!args.Last().StartsWith("-")))
            {
                string dataDir = args.Last() + @"\";

                if (System.IO.File.Exists(dataDir + "train_values.csv"))
                {
                    GenAttrDir = dataDir;
                    return true;
                }
            }

            return (System.IO.File.Exists(GenAttrDir + "train_values.csv"));
        }

        public static string UpdateToCacheDirectory(string fileName)
        {
            return GenAttrInputsDir + @"\" + System.IO.Path.GetFileName(fileName);
        }

        public static int ReadNTest()
        {
            string cacheFileName = UpdateToCacheDirectory(GenAttrTest) + ".nrows.int";

            if (System.IO.File.Exists(cacheFileName))
            {
                return Util.ReadSimpleInt(cacheFileName)[0];
            }

            Util.CreateDirIfNotExists(System.IO.Path.GetDirectoryName(cacheFileName));

            NTestValue = ReadIds(bTrain: false).Length; 

            Util.WriteSimple(cacheFileName, new int[] { NTestValue });

            return NTestValue;
        }

        public static string ReadHeader()
        {
            string fileName = GenAttrLabel;
            string cacheFileName = UpdateToCacheDirectory(fileName) + ".header.string";

            if (System.IO.File.Exists(cacheFileName))
            {
                return System.IO.File.ReadAllText(cacheFileName);
            }

            Util.CreateDirIfNotExists(System.IO.Path.GetDirectoryName(cacheFileName));

            string[] lines = System.IO.File.ReadAllLines(GenAttrLabel);

            System.IO.File.WriteAllText(cacheFileName, lines[0]);

            return lines[0];
        }

        public static int[] ReadLabels()
        {
            string fileName = GenAttrLabel;
            string cacheFileName = UpdateToCacheDirectory(fileName) + ".labels.int";

            if (System.IO.File.Exists(cacheFileName))
            {
                return Util.ReadSimpleInt(cacheFileName);
            }

            Util.CreateDirIfNotExists(System.IO.Path.GetDirectoryName(cacheFileName));

            string[] lines = System.IO.File.ReadAllLines(fileName);
            int[] results = new int[lines.Length - 1];

            for (int i = 1; i < lines.Length; i++)
            {
                string line = lines[i];
                int commasCount = 0;

                for (int j = 0; j < line.Length - 1; j++)
                {
                    if (line[j] == ',')
                    {
                        commasCount++;
                        continue;
                    }
                    if ((line[j] == '1') && (line[j + 1] == '.'))
                    {
                        results[i - 1] = commasCount - 1;
                        break;
                    }
                }
            }

            Util.WriteSimple(cacheFileName, results);

            return results;
        }

        public static int[] ReadLabelsCountsNotSafeForCV()
        {
            string fileName = GenAttrLabel;
            string cacheFileName = UpdateToCacheDirectory(fileName) + ".labels.counts.int";

            if (System.IO.File.Exists(cacheFileName))
            {
                return Util.ReadSimpleInt(cacheFileName);
            }

            Util.CreateDirIfNotExists(System.IO.Path.GetDirectoryName(cacheFileName));

            int[] labels = ReadLabels();

            int[] counts = new int[DataSet.NLabs];

            for (int i = 0; i < labels.Length; i++)
            {
                counts[labels[i]]++;
            }

            Util.WriteSimple(cacheFileName, counts);

            return counts;
        }

        private static string[] ReadColumnsNames(string path, int nStart, int nEnd)
        {
            string cacheFileName = UpdateToCacheDirectory(path) + "." + nStart + "." + nEnd + ".columns.names.string";

            if (System.IO.File.Exists(cacheFileName))
            {
                return System.IO.File.ReadAllLines(cacheFileName);
            }

            Util.CreateDirIfNotExists(System.IO.Path.GetDirectoryName(cacheFileName));

            string[] lines = System.IO.File.ReadAllLines(path);

            string[] columns = lines[0].Split(',').Skip(nStart).Take(nEnd - nStart).ToArray();

            System.IO.File.WriteAllLines(cacheFileName, columns);

            return columns;
        }

        private static ulong[] ReadColumnsULong(string path, int nStart, int nEnd)
        {
            string cacheFileName = UpdateToCacheDirectory(path) + "." + nStart + "." + nEnd + ".columns.ulong";

            if (System.IO.File.Exists(cacheFileName))
            {
                return Util.ReadSimpleULong(cacheFileName);
            }

            Util.CreateDirIfNotExists(System.IO.Path.GetDirectoryName(cacheFileName));

            string[] lines = System.IO.File.ReadAllLines(path);
            ulong[] results = new ulong[lines.Length - 1];

            for (int i = 1; i < lines.Length; i++)
            {
                string[] columns = lines[i].Split(',');
                ulong result = 0;

                for (int j = nEnd - 1; j >= nStart; j--)
                {
                    result = (result << 1) | ((Convert.ToSingle(columns[j]) == 0) ? 0ul : 1ul);
                }

                results[i - 1] = result;
            }

            Util.WriteSimple(cacheFileName, results);

            return results;
        }

        public static string[] ReadStringColumn(string path, int nColumn)
        {
            string cacheFileName = UpdateToCacheDirectory(path) + "." + nColumn + ".string";

            if (System.IO.File.Exists(cacheFileName))
            {
                return System.IO.File.ReadAllLines(cacheFileName);
            }

            Util.CreateDirIfNotExists(System.IO.Path.GetDirectoryName(cacheFileName));

            string[] lines = System.IO.File.ReadAllLines(path);
            string[] results = new string[lines.Length - 1];

            for (int i = 1; i < lines.Length; i++)
            {
                string[] columns = lines[i].Split(',');

                results[i - 1] = columns[nColumn];
            }

            System.IO.File.WriteAllLines(cacheFileName, results);

            return results;
        }

        private static int[] ReadLengths(string path, int nColumn)
        {
            string cacheFileName = UpdateToCacheDirectory(path) + "." + nColumn + ".len.int";

            if (System.IO.File.Exists(cacheFileName))
            {
                return Util.ReadSimpleInt(cacheFileName);
            }

            Util.CreateDirIfNotExists(System.IO.Path.GetDirectoryName(cacheFileName));

            string[] rows = ReadStringColumn(path, nColumn);

            int[] results = rows.Select(v => v.Length).ToArray();

            Util.WriteSimple(cacheFileName, results);

            return results;
        }

        public static int[] ReadNChars(string path, int nColumn, char aChar)
        {
            string cacheFileName = UpdateToCacheDirectory(path) + "." + nColumn + ".base" + aChar + ".int";

            if (System.IO.File.Exists(cacheFileName))
            {
                return Util.ReadSimpleInt(cacheFileName);
            }

            Util.CreateDirIfNotExists(System.IO.Path.GetDirectoryName(cacheFileName));

            string[] rows = ReadStringColumn(path, nColumn);

            int[] results = rows.Select(v => v.Where(w => w == aChar).Count()).ToArray();

            Util.WriteSimple(cacheFileName, results);

            return results;
        }

        public static int[] ReadLengths(bool bTrain)
        {
            return ReadLengths(bTrain ? GenAttrTrain : GenAttrTest, 1);
        }

        public static ulong[] ReadColumnsULong(bool bTrain)
        {
            return ReadColumnsULong(bTrain ? GenAttrTrain : GenAttrTest, 2, 41);
        }

        public static string[] ReadColumnsNames()
        {
            return ReadColumnsNames(GenAttrTrain, 2, 41);
        }

        public static int[] ReadNNum(bool bTrain)
        {
            return ReadNChars(bTrain ? GenAttrTrain : GenAttrTest, 1, 'N');
        }

        public static string[] ReadIds(bool bTrain)
        {
            return ReadStringColumn(bTrain ? GenAttrTrain : GenAttrTest, 0);
        }

        public static string[] ReadSequences(bool bTrain)
        {
            return ReadStringColumn(bTrain ? GenAttrTrain : GenAttrTest, 1);
        }

        public static void SaveResults(double[][] results, bool bSave = true)
        {
            if (!bSave)
            {
                return;
            }

            string[] testIds = ReadIds(bTrain: false);

            string fileName = GenAttrSubmitDir + string.Format("\\ga_{0}.csv", DateTime.Now.ToString(Util.FileDateTimeFormat));

            Util.CreateDirIfNotExists(System.IO.Path.GetDirectoryName(fileName));

            string[] lines = new string[results.Length + 1];

            lines[0] = ReadHeader();

            for (int i = 0; i < results.Length; i++)
            {
                int[] orders = results[i].MyOrderDescending();
                int[] resultsRow = new int[NLabs];

                for (int o = 0; o < TopN; o++)
                {
                    resultsRow[orders[o]] = (o == 0) ? 99 : (TopN - o);
                }

                string line = (testIds[i] + ",0." + string.Join(",0.", resultsRow) + ",").Replace("0.0,", "0.,").Replace("0.99,", "1.,");

                lines[i + 1] = line.Substring(0, line.Length - 1);
            }

            System.IO.File.WriteAllLines(fileName, lines);

            Util.Print("Saved {0} {1}", fileName, DateTime.Now.MyToSecString());
        }
    }
}
