﻿using System;
using GenEngAttr.Data;
using GenEngAttr.Utils;

namespace GenEngAttr.Features
{
    public static class AllFeatures
    {
        public static ulong[] ReadAllBits(bool bTrain)
        {
            string fileNamePrefix = DataSet.GenAttrFeaturesDir + (bTrain ? "train" : "test");
            string cacheFileName = fileNamePrefix + "_allbits.ulong";

            if (System.IO.File.Exists(cacheFileName))
            {
                return Util.ReadSimpleULong(cacheFileName);
            }

            Util.CreateDirIfNotExists(System.IO.Path.GetDirectoryName(cacheFileName));

            ulong[] results = MyBits.GetAllBits(bTrain: bTrain);

            Util.WriteSimple(cacheFileName, results);

            return results;
        }

        public static Tuple<int[], ushort[], int[]> ReadTransform1(bool bTrain, int nGramLength)
        {
            string fileNamePrefix = DataSet.GenAttrFeaturesDir + (bTrain ? "train_t1_" : "test_t1_");
            string fileNamePolymers = fileNamePrefix + "_n" + nGramLength + "_polymers.int";
            string fileNamePlasmids = fileNamePrefix + "_n" + nGramLength + "_plasmids.ushort";
            string fileNameOffsets = fileNamePrefix + "_n" + nGramLength + "_offsets.int";

            if (System.IO.File.Exists(fileNamePolymers))
            {
                return Util.MyTuple(Util.ReadSimpleInt(fileNamePolymers), Util.ReadSimpleUShort(fileNamePlasmids), Util.ReadSimpleInt(fileNameOffsets));
            }

            Util.CreateDirIfNotExists(System.IO.Path.GetDirectoryName(fileNamePolymers));

            string[] sequences = DataSet.ReadSequences(bTrain);
            var (polymers, plasmids, offsets) = Transform1.Transform1Kernel(sequences, nGramLength, nMaxPolymers: 400000000);

            Util.WriteSimple(fileNamePolymers, polymers);
            Util.WriteSimple(fileNamePlasmids, plasmids);
            Util.WriteSimple(fileNameOffsets, offsets);

            return Util.MyTuple(polymers, plasmids, offsets);
        }

        public static Tuple<int[], ushort[]> ReadTransform2(bool bTrain, int nGramLength)
        {
            string fileNamePrefix = DataSet.GenAttrFeaturesDir + (bTrain ? "train_t2_" : "test_t2_");
            string fileNamePolymers = fileNamePrefix + "_n" + nGramLength + "_polymers.int";
            string fileNamePlasmids = fileNamePrefix + "_n" + nGramLength + "_plasmids.ushort";

            if (System.IO.File.Exists(fileNamePlasmids))
            {
                return Util.MyTuple(Util.ReadSimpleInt(fileNamePolymers), Util.ReadSimpleUShort(fileNamePlasmids));
            }

            Util.CreateDirIfNotExists(System.IO.Path.GetDirectoryName(fileNamePolymers));

            var (polymers, plasmids, offsets) = ReadTransform1(bTrain, nGramLength);

            var (t2polymers, t2plasmids) = Transform2.Transform2Kernel(polymers, plasmids);

            Util.WriteSimple(fileNamePolymers, t2polymers);
            Util.WriteSimple(fileNamePlasmids, t2plasmids);

            return Util.MyTuple(t2polymers, t2plasmids);
        }

        public static Tuple<int[], int[], int[], float[]> ReadTransform3(int nGramLength)
        {
            string fileNamePrefix = DataSet.GenAttrFeaturesDir + "train_t3_";
            string fileNamePolymers = fileNamePrefix + "_n" + nGramLength + "_polymers.int";
            string fileNameLabs = fileNamePrefix + "_n" + nGramLength + "_labs.int";
            string fileNameCounts = fileNamePrefix + "_n" + nGramLength + "_counts.int";
            string fileNameShares = fileNamePrefix + "_n" + nGramLength + "_shares.float";

            if (System.IO.File.Exists(fileNameShares))
            {
                return Util.MyTuple(Util.ReadSimpleInt(fileNamePolymers), Util.ReadSimpleInt(fileNameLabs),
                    Util.ReadSimpleInt(fileNameCounts), Util.ReadSimpleFloat(fileNameShares));
            }

            Util.CreateDirIfNotExists(System.IO.Path.GetDirectoryName(fileNameLabs));

            var (polymers, plasmids) = ReadTransform2(bTrain: true, nGramLength: nGramLength);

            var labels = DataSet.ReadLabels();

            var (t3polymers, t3labs, t3counts, t3shares) = Transform3.Transform3Kernel(polymers, plasmids, labels);

            Util.WriteSimple(fileNamePolymers, t3polymers);
            Util.WriteSimple(fileNameLabs, t3labs);
            Util.WriteSimple(fileNameCounts, t3counts);
            Util.WriteSimple(fileNameShares, t3shares);

            return Util.MyTuple(t3polymers, t3labs, t3counts, t3shares);
        }

        public static ushort[] ReadPersonalPlasmid(bool bTrain, int nGramLength, int iPlasmid)
        {
            string fileNamePrefix = DataSet.GenAttrFeaturesDir + (bTrain ? "train_personalPlasmid_" : "test_personalPlasmid_");
            string cacheFileName = fileNamePrefix + iPlasmid + "_n" + nGramLength + ".ushort";

            if (System.IO.File.Exists(cacheFileName))
            {
                return Util.ReadSimpleUShort(cacheFileName);
            }

            Util.CreateDirIfNotExists(System.IO.Path.GetDirectoryName(cacheFileName));

            ushort[][] personals = PersonalPlasmid.SimpleSimPersonalKernel(bTrain, nGramLength, new int[] { iPlasmid });

            ushort[] results = personals[0];

            Util.WriteSimple(cacheFileName, results);

            return results;
        }
    }
}
