﻿using System;
using System.Collections.Generic;
using GenEngAttr.Data;
using GenEngAttr.Utils;

namespace GenEngAttr.Features
{
    // k-mer transforms here
    static class Transform1 
    {
        public const int NMaxPolymers = 40000000;

        // returns polymers, plasmids, offsets; offsets are not actually used anywhere
        public static Tuple<int[], ushort[], int[]> Transform1Kernel(string[] sequences, int nGramLength, int nMaxPolymers = NMaxPolymers)
        {
            List<int> polymers = new List<int>(nMaxPolymers);
            List<ushort> plasmids = new List<ushort>(nMaxPolymers);
            List<int> offsets = new List<int>(nMaxPolymers);

            Util.Print("start of Transform1Kernel {0} {1}", nGramLength, DateTime.Now.MyToSecString());

            for (int i = 0; i < sequences.Length; i++)
            {
                Transform(polymers, plasmids, offsets, i, sequences[i], nGramLength);
            }

            Util.Print("end of Transform1 {0} {1} {2} {3}", polymers.Count, plasmids.Count, offsets.Count, DateTime.Now.MyToSecString());

            int[] orders = polymers.MyOrder();

            Util.Print("end of MyOrder {0} {1}", orders.Length, DateTime.Now.MyToSecString());

            int[] aPolymers = new int[orders.Length];
            ushort[] aPlasmids = new ushort[orders.Length];
            int[] aOffsets = new int[orders.Length];

            for (int o = 0; o < orders.Length; o++)
            {
                int i = orders[o];

                aPolymers[o] = polymers[i];
            }

            for (int o = 0; o < orders.Length; o++)
            {
                int i = orders[o];

                aPlasmids[o] = plasmids[i];
            }

            for (int o = 0; o < orders.Length; o++)
            {
                int i = orders[o];

                aOffsets[o] = offsets[i];
            }

            Util.Print("end of Re-Arrange {0}", DateTime.Now.MyToSecString());

            return Util.MyTuple(aPolymers, aPlasmids, aOffsets);
        }

        public static void Transform(List<int> polymers, List<ushort> plasmids, List<int> offsets, int plasmid, string sequence, int nGramLength)
        {
            if (nGramLength > 16)
            {
                Transform32(polymers, plasmids, offsets, plasmid, sequence, nGramLength);
                return;
            }

            int polymer = 0;
            int length = 0;
            int mask = (int)(((uint)0xFFFFFFFF) >> (32 - 2 * nGramLength));

            for (int i = 0; i < sequence.Length; i++)
            {
                int nucleo = GAUtil.GetNum(sequence[i]);

                if (nucleo == 5)
                {
                    polymer = 0;
                    length = 0;
                    continue;
                }

                polymer = ((polymer << 2) + nucleo) & mask;
                length++;

                if (length >= nGramLength)
                {
                    polymers.Add(polymer);
                    plasmids.Add((ushort)plasmid);
                    offsets.Add(i - nGramLength + 1);
                }
            }
        }

        public static void Transform32(List<int> polymers, List<ushort> plasmids, List<int> offsets, int plasmid, string sequence, int nGramLength)
        {
            int polymerA = 0;
            int polymerB = 0;
            int polymerC = 0;
            int length = 0;
            int shiftB = 2 * ((nGramLength + 1) / 2);
            int maskA = (int)(((uint)0xFFFFFFFF) >> (32 - 2 * ((nGramLength + 1) / 2)));
            int maskB = (int)(((uint)0xFFFFFFFF) >> (32 - 2 * ((nGramLength) / 2)));

            for (int i = 0; i < sequence.Length; i++)
            {
                int nucleo = GAUtil.GetNum(sequence[i]);

                if (nucleo == 5)
                {
                    polymerA = 0;
                    polymerB = 0;
                    length = 0;
                    continue;
                }

                polymerC = polymerB;
                polymerB = polymerA;
                polymerA = polymerC;
                polymerA = ((polymerA << 2) + nucleo);

                length++;

                if (length >= nGramLength)
                {
                    polymerC = (polymerA & maskA) | ((polymerB & maskB) << shiftB);
                    polymers.Add(polymerC);
                    plasmids.Add((ushort)plasmid);
                    offsets.Add(i - nGramLength + 1);
                }
            }
        }

        public static int GetPolymer(string sequence, int offset, int nGramLength)
        {
            int polymer = 0;

            for (int i = offset; i < offset + nGramLength; i++)
            {
                int nucleo = GAUtil.GetNum(sequence[i]);
                polymer = (polymer << 2) + nucleo;
            }

            return polymer;
        }

    }
}
