﻿using System;
using System.Collections.Generic;
using GenEngAttr.Data;
using GenEngAttr.Utils;

namespace GenEngAttr.Features
{
    class Transform2
    {
        public static Tuple<int[], ushort[]> Transform2Kernel(int[] polymers, ushort[] plasmids)
        {
            List<int> t2polymers = new List<int>(polymers.Length);
            List<ushort> t2plasmids = new List<ushort>(plasmids.Length);
            List<ushort> tempPlasmids = new List<ushort>(DataSet.NTrain);

            Util.Print("start of Transform2Kernel {0}", DateTime.Now.MyToSecString());

            for (int i = 0; i < polymers.Length; i++)
            {
                tempPlasmids.Add(plasmids[i]);

                if ((i == polymers.Length - 1) || (polymers[i] != polymers[i + 1]))
                {
                    tempPlasmids.Sort();

                    for (int j = 0; j < tempPlasmids.Count; j++)
                    {
                        if ((j == tempPlasmids.Count - 1) || (tempPlasmids[j] != tempPlasmids[j + 1]))
                        {
                            t2polymers.Add(polymers[i]);
                            t2plasmids.Add(tempPlasmids[j]);
                        }
                    }

                    tempPlasmids.Clear();
                }

                //if ((i & 0xFFFFF) == 0)
                //{
                //    Util.Print("Transform2 {0}", i);
                //}
            }

            Util.Print("end of Transform2Kernel {0}", DateTime.Now.MyToSecString());

            return Util.MyTuple(t2polymers.ToArray(), t2plasmids.ToArray());
        }

        public static Tuple<int[], ushort[]> Transform2GetOnlyDisables(int[] t2polymers, ushort[] t2plasmids, bool[] disables)
        {
            ushort[] newPlasmidsIndex = new ushort[disables.Length];
            ushort uNewPlasmidsIndex = 0;

            for (int i = 0; i < disables.Length; i++)
            {
                if (disables[i])
                {
                    newPlasmidsIndex[i] = uNewPlasmidsIndex;
                    uNewPlasmidsIndex++;
                }
            }

            int n = 0;

            for (int i = 0; i < t2plasmids.Length; i++)
            {
                n += (disables[t2plasmids[i]] ? 1 : 0);
            }

            int[] resPolymers = new int[n];
            ushort[] resPlasmids = new ushort[n];

            int offset = 0;

            for (int i = 0; i < t2plasmids.Length; i++)
            {
                if (disables[t2plasmids[i]])
                {
                    resPolymers[offset] = t2polymers[i];
                    resPlasmids[offset] = newPlasmidsIndex[t2plasmids[i]];
                    offset++;
                }
            }

            return Util.MyTuple(resPolymers, resPlasmids);
        }

        public static Tuple<int[], int[]> Transform2GetOnlyDisables(int[] t2polymers, int[] t2plasmids, bool[] disables)
        {
            int[] newPlasmidsIndex = new int[disables.Length];
            int uNewPlasmidsIndex = 0;

            for (int i = 0; i < disables.Length; i++)
            {
                if (disables[i])
                {
                    newPlasmidsIndex[i] = uNewPlasmidsIndex;
                    uNewPlasmidsIndex++;
                }
            }

            int n = 0;

            for (int i = 0; i < t2plasmids.Length; i++)
            {
                n += (disables[t2plasmids[i]] ? 1 : 0);
            }

            int[] resPolymers = new int[n];
            int[] resPlasmids = new int[n];

            int offset = 0;

            for (int i = 0; i < t2plasmids.Length; i++)
            {
                if (disables[t2plasmids[i]])
                {
                    resPolymers[offset] = t2polymers[i];
                    resPlasmids[offset] = newPlasmidsIndex[t2plasmids[i]];
                    offset++;
                }
            }

            return Util.MyTuple(resPolymers, resPlasmids);
        }

        public static Tuple<int[], int[]> GetJoinTransform2Kernel(int nGramLength)
        {
            var (trainPolymers, trainPlasmids) = AllFeatures.ReadTransform2(bTrain: true, nGramLength: nGramLength);
            var (testPolymers, testPlasmids) = AllFeatures.ReadTransform2(bTrain: false, nGramLength: nGramLength);

            Util.Print("start of GetJoinTransform2Kernel {0}", DateTime.Now.MyToSecString());

            int totalCount = trainPolymers.Length + testPolymers.Length;
            int testPlasmidAdd = DataSet.NTrain;

            int[] trainTestPolymers = new int[totalCount];
            int[] trainTestPlasmids = new int[totalCount];

            int trainTestOffset = 0;
            int trainOffset = 0;
            int testOffset = 0;

            while (trainTestOffset < totalCount)
            {
                // always trainPlasmids < testPlasmids and t2 is also sorted by plasmids
                if ((trainOffset < trainPolymers.Length) && (trainPolymers[trainOffset] <= testPolymers[testOffset]))
                {
                    trainTestPolymers[trainTestOffset] = trainPolymers[trainOffset];
                    trainTestPlasmids[trainTestOffset] = trainPlasmids[trainOffset];
                    trainTestOffset++;
                    trainOffset++;
                }
                else
                {
                    trainTestPolymers[trainTestOffset] = testPolymers[testOffset];
                    trainTestPlasmids[trainTestOffset] = testPlasmids[testOffset] + testPlasmidAdd;
                    trainTestOffset++;
                    testOffset++;
                }
            }

            Util.Print("end of GetJoinTransform2Kernel {0}", DateTime.Now.MyToSecString());

            return Util.MyTuple(trainTestPolymers, trainTestPlasmids);
        }
    }
}
