﻿using System;
using System.Collections.Generic;
using GenEngAttr.Data;
using GenEngAttr.Utils;

namespace GenEngAttr.Models
{
    // applying and caching models
    class AllModels
    {
        public static void ApplySoftMasks(double[][] resLabs, double pow = 0.5) // hamming weights distance model
        {
            Util.Print("ApplySoftMasks started {0}", DateTime.Now.MyToSecString());
            SoftMasks.ApplySoftMasks(resLabs, pow);
        }

        public static void ApplyHighSoftMasks(double[][] resLabs, int[][] top10, double beta = 0.02) // hamming weights distance model on top of others models results
        {
            Util.Print("ApplyHighSoftMasks started {0}", DateTime.Now.MyToSecString());
            HighSoftMasks.ApplyHighSoftMasks(resLabs, top10, beta);
        }

        public static double[][] ApplyHighPlasmidsClusters(double[][] resLabsSrc, int plGramLength, double beta) // clustering on test plasmids similarity matrixes
        {
            Util.Print("ApplyHighPlasmidsClusters started {0}", DateTime.Now.MyToSecString());
            return HighPlasmidsClusters.ApplyHighPlasmidsClusters(resLabsSrc, plGramLength, beta);
        }

        public static double[][] ReadPlasmids(int plGramLength) // based on plasmid to plasmid similarity
        {
            Util.Print("ReadPlasmids started {0}", DateTime.Now.MyToSecString());
            return CacheOutModel(PlasmidsModels.GetPlasmids, plGramLength);
        }

        public static double[][] ReadPlasmidsHyb1(int plGramLength) // based on plasmid to plasmid similarity, but no ngrams of top1 lab
        {
            Util.Print("ReadPlasmidsHyb1 started {0}", DateTime.Now.MyToSecString());
            return CacheOutModel(PlasmidsModels.GetPlasmidsHyb1, plGramLength);
        }

        public static double[][] ReadT3(int t3GramLength) // plasmid to lab similarity model
        {
            Util.Print("ReadT3 started {0}", DateTime.Now.MyToSecString());
            return CacheOutModel(T3Models.GetT3, t3GramLength);
        }

        public static double[][] ReadT3Hyb1(int t3GramLength) // plasmid to lab similarity model, but no ngrams of top1 lab
        {
            Util.Print("ReadT3Hyb1 started {0}", DateTime.Now.MyToSecString());
            return CacheOutModel(T3Models.GetT3Hyb1, t3GramLength);
        }

        public static double[][] ReadHighT3(int t3GramLength) // plasmid to lab similarity model when test plasmids are labeled by top1 lab of Wave9462 
        {
            Util.Print("ReadHighT3 started {0}", DateTime.Now.MyToSecString());
            return CacheOutModel(HighT3.GetHighT3, t3GramLength);
        }

        public static double[][] ReadPartialBayes() // bayes on particular bits mul model
        {
            Util.Print("ReadPartialBayes started {0}", DateTime.Now.MyToSecString());
            return CacheOutModel(PartialBayes.GetPartialBayes);
        }

        public static double[][] ReadHighPartialBayes() // particular bits mul model when test plasmids are labeled by top1 lab of Wave9462 
        {
            Util.Print("ReadHighPartialBayes started {0}", DateTime.Now.MyToSecString());
            return CacheOutModel(HighPartialBayes.GetHighPartialBayes);
        }

        public static double[][] ReadWeightBayes16() // weighted bayes 
        {
            Util.Print("ReadWeightBayes16 started {0}", DateTime.Now.MyToSecString());
            return CacheOutModel(WeightBayes.GetWeightBayes16);
        }

        public static double[][] ReadWaveA(int plGramLength) // (genetic + bayes16) * partial bayes
        {
            Util.Print("ReadWaveA started {0}", DateTime.Now.MyToSecString());
            return CacheOutModel(WavesModels.GetWaveA, plGramLength);
        }

        public static double[][] ReadWaveB(int plGramLength) // wave9396 + improve masks
        {
            Util.Print("ReadWaveB started {0}", DateTime.Now.MyToSecString());
            return CacheOutModel(WavesModels.GetWaveB, plGramLength);
        }

        public static ushort[][] ReadSimPlasmids(int nGramLength) // plasmid to plasmid similarity matrix
        {
            Util.Print("ReadSimPlasmids started {0}", DateTime.Now.MyToSecString());
            return CacheOutPlasmids(PlasmidsSim.GetPlasmidsSim, nGramLength);
        }

        public static ushort[][] ReadSimPlasmidsHyb1(int nGramLength) // plasmid to plasmid similarity matrix, but no ngrams of top1 lab
        {
            Util.Print("ReadSimPlasmidsHyb1 started {0}", DateTime.Now.MyToSecString());
            return CacheOutPlasmids(PlasmidsSim.GetPlasmidsSimHyb1, nGramLength);
        }

        public static ushort[][] ReadSimHighPlasmids(int nGramLength) // test plasmid to test plasmid similarity matrix
        {
            Util.Print("ReadSimHighPlasmids started {0}", DateTime.Now.MyToSecString());
            return CacheOutPlasmids(HighPlasmidsSim.GetHighPlasmidsSim, nGramLength, DataSet.NTest);
        }

        public static int[] ReadHighClusterIds(int plGramLength) // clustering on test plasmid based plasmid to plasmid similarity
        {
            Util.Print("ReadHighClusterIds started {0}", DateTime.Now.MyToSecString());
            return CacheArray(HighPlasmidsClusters.GetHighClusterIds, plGramLength);
        }

        public static int[][] ReadWaveCTop10(int plGramLength) // WaveB + clustering 
        {
            Util.Print("ReadWaveCTop10 started {0}", DateTime.Now.MyToSecString());
            return CacheOnDirectoryInt(WavesModels.GetWaveCTop10, plGramLength, DataSet.GenAttrOutModelsDir, 10);
        }

        public static int[][] ReadWaveDTop10(int plGramLength) // WaveA + high masks + high clustering 
        {
            Util.Print("ReadWaveDTop10 started {0}", DateTime.Now.MyToSecString());
            return CacheOnDirectoryInt(WavesModels.GetWaveDTop10, plGramLength, DataSet.GenAttrOutModelsDir, 10);
        }

        public static int[] CacheArray<T1>(Func<T1, int[]> CreationFunction, T1 t1)
        {
            string paramString = "." + t1.ToString();
            string fileName = DataSet.GenAttrOutModelsDir + CreationFunction.Method.Name + paramString + ".int";

            if (System.IO.File.Exists(fileName))
            {
                int[] resSimple = Util.ReadSimpleInt(fileName);

                return resSimple;
            }

            Util.CreateDirIfNotExists(System.IO.Path.GetDirectoryName(fileName));

            int[] res = CreationFunction(t1);

            Util.WriteSimple(fileName, res);

            return res;
        }

        public static int[][] CacheOnDirectoryInt<T1>(Func<T1, int[][]> CreationFunction, T1 t1, string directory, int colLength)
        {
            string paramString = "." + t1.ToString();
            string fileName = directory + CreationFunction.Method.Name + paramString + ".int";

            if (System.IO.File.Exists(fileName))
            {
                int[] resSimple = Util.ReadSimpleInt(fileName);
                int rowLength = resSimple.Length / colLength;

                return Util.VectorToMatrix(resSimple, rowLength);
            }

            Util.CreateDirIfNotExists(System.IO.Path.GetDirectoryName(fileName));

            int[][] trainRes = CreationFunction(t1);
            int[] res = Util.MatrixToVector(trainRes);

            Util.WriteSimple(fileName, res);

            return trainRes;
        }

        private static double[][] CacheOutModel(Func<double[][]> CreationFunction)
        {
            string paramString = "";
            string fileName = DataSet.GenAttrOutModelsDir + CreationFunction.Method.Name + paramString + ".double";
            List<double[]> res = new List<double[]>();

            if (System.IO.File.Exists(fileName + 0))
            {
                int i = 0;

                while (System.IO.File.Exists(fileName + i))
                {
                    double[] resSimple = Util.ReadSimpleDouble(fileName + i);

                    res.AddRange(Util.VectorToMatrix(resSimple, DataSet.NLabs));

                    i++;
                }

                return res.ToArray();
            }

            Util.CreateDirIfNotExists(System.IO.Path.GetDirectoryName(fileName));

            double[][] trainRes = CreationFunction();

            int nMaxElements = 134217000;// 134217728 2^27 1Gb
            int nMaxRows = (nMaxElements + trainRes[0].Length - 1) / trainRes[0].Length;
            int nPieces = (trainRes.Length + nMaxRows - 1) / nMaxRows;

            for (int i = 0; i < nPieces; i++)
            {
                int iStart = i * nMaxRows;
                int iEnd = Math.Min(trainRes.Length, iStart + nMaxRows);
                double[][] partTrainRes = new double[iEnd - iStart][];

                for (int j = iStart; j < iEnd; j++)
                {
                    partTrainRes[j - iStart] = trainRes[j];
                }

                double[] partVector = Util.MatrixToVector(partTrainRes);

                Util.WriteSimple(fileName + i, partVector);
            }

            return trainRes;
        }

        private static double[][] CacheOutModel<T1>(Func<T1, double[][]> CreationFunction, T1 t1)
        {
            string paramString = "." + t1.ToString();
            string fileName = DataSet.GenAttrOutModelsDir + CreationFunction.Method.Name + paramString + ".double";
            List<double[]> res = new List<double[]>();

            if (System.IO.File.Exists(fileName + 0))
            {
                int i = 0;

                while (System.IO.File.Exists(fileName + i))
                {
                    double[] resSimple = Util.ReadSimpleDouble(fileName + i);

                    res.AddRange(Util.VectorToMatrix(resSimple, DataSet.NLabs));

                    i++;
                }

                return res.ToArray();
            }

            Util.CreateDirIfNotExists(System.IO.Path.GetDirectoryName(fileName));

            double[][] trainRes = CreationFunction(t1);

            int nMaxElements = 134217000;// 134217728 2^27 1Gb
            int nMaxRows = (nMaxElements + trainRes[0].Length - 1) / trainRes[0].Length;
            int nPieces = (trainRes.Length + nMaxRows - 1) / nMaxRows;

            for (int i = 0; i < nPieces; i++)
            {
                int iStart = i * nMaxRows;
                int iEnd = Math.Min(trainRes.Length, iStart + nMaxRows);
                double[][] partTrainRes = new double[iEnd - iStart][];

                for (int j = iStart; j < iEnd; j++)
                {
                    partTrainRes[j - iStart] = trainRes[j];
                }

                double[] partVector = Util.MatrixToVector(partTrainRes);

                Util.WriteSimple(fileName + i, partVector);
            }

            return trainRes;
        }

        private static ushort[][] CacheOutPlasmids<T1>(Func<T1, ushort[][]> CreationFunction, T1 t1, int rowLength = DataSet.NTrain)
        {
            string paramString = "." + t1.ToString();
            string fileName = DataSet.GenAttrOutModelsDir + CreationFunction.Method.Name + paramString + ".ushort";
            List<ushort[]> res = new List<ushort[]>();

            if (System.IO.File.Exists(fileName + 0))
            {
                int i = 0;

                while (System.IO.File.Exists(fileName + i))
                {
                    ushort[] resSimple = Util.ReadSimpleUShort(fileName + i);

                    res.AddRange(Util.VectorToMatrix(resSimple, rowLength));

                    i++;
                }

                return res.ToArray();
            }

            Util.CreateDirIfNotExists(System.IO.Path.GetDirectoryName(fileName));

            ushort[][] trainRes = CreationFunction(t1);

            int nMaxElements = 536870000; //536870912; 2^29 1Gb
            int nMaxRows = (nMaxElements + trainRes[0].Length - 1) / trainRes[0].Length;
            int nPieces = (trainRes.Length + nMaxRows - 1) / nMaxRows;

            for (int i = 0; i < nPieces; i++)
            {
                int iStart = i * nMaxRows;
                int iEnd = Math.Min(trainRes.Length, iStart + nMaxRows);
                ushort[][] partTrainRes = new ushort[iEnd - iStart][];

                for (int j = iStart; j < iEnd; j++)
                {
                    partTrainRes[j - iStart] = trainRes[j];
                }

                ushort[] partVector = Util.MatrixToVector(partTrainRes);

                Util.WriteSimple(fileName + i, partVector);
            }

            return trainRes;
        }
    }
}
