﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace GenEngAttr.Models
{
    class HighPlasmidsClusters
    {
        public static double[][] ApplyHighPlasmidsClusters(double[][] resLabsSrc, int plGramLength, double beta)
        {
            int[] clusterIdsAr = AllModels.ReadHighClusterIds(plGramLength);
            double[][] resLabsCluster = resLabsSrc.Select(v => v.ToArray()).ToArray();

            int offset = 0;

            while (offset < clusterIdsAr.Length)
            {
                int i = clusterIdsAr[offset++];
                int count = clusterIdsAr[offset++];

                if (count > 0)
                {
                    for (int j = 0; j < resLabsCluster[i].Length; j++)
                    {
                        resLabsCluster[i][j] *= beta;
                    }

                    for (int k = 0; k < count; k++)
                    {
                        int ids = clusterIdsAr[offset++];

                        for (int j = 0; j < resLabsCluster[i].Length; j++)
                        {
                            resLabsCluster[i][j] += resLabsSrc[ids][j];
                        }
                    }

                    double revDivider = (count + beta);

                    for (int j = 0; j < resLabsCluster[i].Length; j++)
                    {
                        resLabsCluster[i][j] *= revDivider;
                    }
                }
            }

            return resLabsCluster;
        }

        public static int[] GetHighClusterIds(int plGramLength) 
        {
            var highPlasmids = AllModels.ReadSimHighPlasmids(plGramLength);
            var resLabsPl = AllModels.ReadPlasmids(plGramLength);
            var plasmidMaxes = resLabsPl.Select(v => v.Max()).ToArray();

            int[] maxes = new int[highPlasmids.Length];

            for (int i = 0; i < highPlasmids.Length; i++)
            {
                maxes[i] = 0;

                highPlasmids[i][i] = 0;

                for (int j = 0; j < highPlasmids[i].Length; j++)
                {
                    maxes[i] = Math.Max(maxes[i], highPlasmids[i][j]);
                }
            }

            List<int> clusterIdsLs = new List<int>();

            for (int i = 0; i < highPlasmids.Length; i++)
            {
                if (maxes[i] > plasmidMaxes[i])
                {
                    List<int> clusterIds = new List<int>();

                    for (int j = 0; j < highPlasmids.Length; j++)
                    {
                        if (highPlasmids[i][j] == maxes[i])
                        {
                            clusterIds.Add(j);
                        }
                    }

                    clusterIdsLs.Add(i);
                    clusterIdsLs.Add(clusterIds.Count);
                    clusterIdsLs.AddRange(clusterIds);
                }
            }

            return clusterIdsLs.ToArray();
        }
    }
}
