﻿using System;
using System.Collections.Generic;
using System.Linq;
using GenEngAttr.Data;
using GenEngAttr.Utils;

namespace GenEngAttr.Models
{
    // soft mask model
    public static class HighSoftMasks
    {
        public static void ApplyHighSoftMasks(double[][] resLabs, int[][] top10, double beta = 0.02)
        {
            int[] labels = DataSet.ReadLabels();
            ulong[] trainColumnsULong = DataSet.ReadColumnsULong(bTrain: true);
            ulong[] testColumnsULong = DataSet.ReadColumnsULong(bTrain: false);
            Dictionary<ulong, Dictionary<int, double>> bitDictionary = new Dictionary<ulong, Dictionary<int, double>>();
            Dictionary<ulong, Tuple<int, double>> bitDictionaryCounts = new Dictionary<ulong, Tuple<int, double>>();
            Dictionary<ulong, KeyValuePair<int, double>[]> forceLabsDictionary = new Dictionary<ulong, KeyValuePair<int, double>[]>();

            for (int i = 0; i < trainColumnsULong.Length; i++)
            {
                if (!bitDictionary.ContainsKey(trainColumnsULong[i]))
                {
                    bitDictionary[trainColumnsULong[i]] = new Dictionary<int, double>();
                }

                Dictionary<int, double> labDictionary = bitDictionary[trainColumnsULong[i]];

                if (labDictionary.ContainsKey(labels[i]))
                {
                    labDictionary[labels[i]] += 1;
                }
                else
                {
                    labDictionary[labels[i]] = 1;
                }
            }

            for (int i = 0; i < testColumnsULong.Length; i++)
            {
                if (!bitDictionary.ContainsKey(testColumnsULong[i]))
                {
                    bitDictionary[testColumnsULong[i]] = new Dictionary<int, double>();
                }

                Dictionary<int, double> labDictionary = bitDictionary[testColumnsULong[i]];

                if (labDictionary.ContainsKey(top10[0][i]))
                {
                    labDictionary[top10[0][i]] += 1;
                }
                else
                {
                    labDictionary[top10[0][i]] = 1;
                }
            }

            foreach (var keyValue in bitDictionary)
            {
                ulong bits = keyValue.Key;
                Dictionary<int, double> labDictionary = keyValue.Value;

                int labCount = labDictionary.Count();
                double totalCount = labDictionary.Select(v => v.Value).Sum();

                bitDictionaryCounts.Add(bits, Util.MyTuple(labCount, totalCount));

                if ((totalCount > 0))
                {
                    int lab = labDictionary.First().Key;

                    var labs = labDictionary.OrderByDescending(v => v.Value).ToArray();

                    forceLabsDictionary[bits] = labs;
                }
            }

            for (int i = 0; i < resLabs.Length; i++)
            {
                ulong bits = testColumnsULong[i];
                var labs = forceLabsDictionary[bits];
                double alpha = Math.Pow(labs.Length, 0.5);

                for (int j = 0; j < labs.Length; j++)
                {
                    resLabs[i][labs[j].Key] += 1d / (alpha + beta * labs[0].Value / labs[j].Value);
                }
            }
        }
    }
}
