﻿using System;
using System.Collections.Generic;
using System.Linq;
using GenEngAttr.Data;
using GenEngAttr.Features;
using GenEngAttr.Utils;

namespace GenEngAttr.Models
{
    // T3 model on top of other models
    public static class HighT3
    {
        // same as T3Kernel just with int[] testPlasmids
        public static double[][] T3KernelInt(int trainLength, int testLength,
            int[] trainPolymers, int[] trainLabs, float[] weights, int[] testPolymers, int[] testPlasmids, int[][] devSetTops)
        {
            double[][] resLabs = new double[testLength][];

            for (int i = 0; i < resLabs.Length; i++)
            {
                resLabs[i] = new double[trainLength];
            }

            int iStart = 0;
            int jStart = 0;

            for (int i = 0; i < testPolymers.Length; i++)
            {
                if ((i == testPolymers.Length - 1) || (testPolymers[i] != testPolymers[i + 1]))
                {
                    while ((jStart < trainPolymers.Length) && (trainPolymers[jStart] < testPolymers[i])) jStart++;

                    if (jStart == trainPolymers.Length)
                    {
                        break;
                    }

                    if (trainPolymers[jStart] != testPolymers[i])
                    {
                        iStart = i + 1;
                        continue;
                    }

                    int jEnd = jStart;

                    while ((jEnd < trainPolymers.Length) && (trainPolymers[jEnd] == testPolymers[i])) jEnd++;

                    for (int ik = iStart; ik <= i; ik++)
                    {
                        bool bNewPolymer = true;

                        for (int l = 0; bNewPolymer && (l < devSetTops.Length); l++)
                        {
                            for (int j = jStart; j < jEnd; j++)
                            {
                                if (trainLabs[j] == devSetTops[l][testPlasmids[ik]])
                                {
                                    bNewPolymer = false;
                                    break;
                                }
                            }
                        }

                        if (bNewPolymer)
                        {
                            for (int j = jStart; j < jEnd; j++)
                            {
                                resLabs[testPlasmids[ik]][trainLabs[j]] += weights[j];
                            }
                        }
                    }

                    iStart = i + 1;
                    jStart = jEnd;
                }

                if ((i & 0xFFFFF) == 0)
                {
                    //Util.Print("T3KernelInt {0}/{1} {2}", i, testPolymers.Length, DateTime.Now.MyToLongString());
                }
            }

            return resLabs;
        }

        public static double[][] GetHighT3HybMT(int t3GramLength, int[][] resLabsTops, int[][] top10)
        {
            var (labels, devSet) = GAUtil.GetOutHighLabelsAndDevSet();

            bool[] trainTestDisables = Util.GetDisables(labels.Length, devSet);

            var (trainTestPolymers, trainTestPlasmids) = Transform2.GetJoinTransform2Kernel(t3GramLength);

            int[] newLabels = labels.ToArray();

            for (int i = 0; i < devSet.Length; i++)
            {
                newLabels[devSet[i]] = top10[0][i];
            }

            var (trainT3Polymers, trainLabs, t3Counts, trainShares) = Transform3.Transform3KernelCVDevInt(trainTestPolymers, trainTestPlasmids, newLabels, trainTestDisables);

            double[][][] res = new double[24][][];

            Util.MyParallelFor(6, 0, res.Length, (i) =>
            {
                var (partDisables, nDisable) = GAUtil.GetPartDisables(trainTestDisables, i, res.Length);

                var (devt2Polymers, devt2Plasmids) = Transform2.Transform2GetOnlyDisables(trainTestPolymers, trainTestPlasmids, partDisables);

                int[][] devSetTops = GAUtil.GetDevSetTops(resLabsTops, devSet, partDisables);

                res[i] = T3KernelInt(DataSet.NLabs, nDisable, trainT3Polymers, trainLabs, trainShares, devt2Polymers, devt2Plasmids, devSetTops);
            });

            List<double[]> resLs = new List<double[]>();

            for (int i = 0; i < res.Length; i++)
            {
                resLs.AddRange(res[i]);
            }

            return resLs.ToArray();
        }

        public static double[][] GetHighT3(int t3GramLength)
        {
            int[][] top10 = AllModels.ReadWaveDTop10(plGramLength: 15);

            var resLabsHyb = GetHighT3HybMT(t3GramLength, new int[][] { }, top10);

            return resLabsHyb;
        }
    }
}
