﻿using System;
using System.Linq;
using GenEngAttr.Data;
using GenEngAttr.Features;
using GenEngAttr.Utils;

namespace GenEngAttr.Models
{
    class PartialBayes
    {
        public static double[][] CalcPartialBayes(int[] labels, ulong[] trainBits, ulong[] testBits, int[] kds, bool bOne = true)
        {
            float alphaRegulator = 1;
            double[][] resLabs = new double[DataSet.NTest][];
            float[] totalLabsCount = new float[DataSet.NLabs];
            float[][] labsCounters = Util.InitArray(0, DataSet.NLabs).Select(v => new float[kds.Length]).ToArray();
            float[][] oneVals = Util.InitArray(0, DataSet.NLabs).Select(v => new float[kds.Length]).ToArray();
            float[][] zeroVals = Util.InitArray(0, DataSet.NLabs).Select(v => new float[kds.Length]).ToArray();
            float[] columnsCounters = new float[kds.Length];
            float totalCount = 0;

            for (int i = 0; i < trainBits.Length; i++)
            {
                int thisLab = labels[i];

                totalLabsCount[thisLab]++;

                for (int j = 0; j < kds.Length; j++)
                {
                    int value = (int)((trainBits[i] >> kds[j]) & (1UL));

                    labsCounters[thisLab][j] += value;
                    columnsCounters[j] += value;
                }

                totalCount++;
            }

            for (int j = 0; j < DataSet.NLabs; j++)
            {
                float divider = (1f / Math.Max(1, totalLabsCount[j]));

                for (int k = 0; k < kds.Length; k++)
                {
                    oneVals[j][k] = (labsCounters[j][k] + alphaRegulator) / (columnsCounters[k] + alphaRegulator) * divider;
                    zeroVals[j][k] = ((totalLabsCount[j] - labsCounters[j][k]) + alphaRegulator) / ((totalCount - columnsCounters[k]) + alphaRegulator) * divider;
                }
            }

            for (int i = 0; i < testBits.Length; i++)
            {
                resLabs[i] = new double[DataSet.NLabs];

                for (int j = 0; j < resLabs[i].Length; j++)
                {
                    resLabs[i][j] = bOne ? 1f : Math.Max(1, totalLabsCount[j]);

                    for (int k = 0; k < kds.Length; k++)
                    {
                        if (((testBits[i] >> kds[k]) & (1UL)) == 1)
                        {
                            resLabs[i][j] *= oneVals[j][k];
                        }
                        else
                        {
                            resLabs[i][j] *= zeroVals[j][k];
                        }
                    }
                }
            }

            return resLabs;
        }

        public static double[][] GetPartialBayes()
        {
            int[] labels = DataSet.ReadLabels();
            ulong[] trainAllBits = AllFeatures.ReadAllBits(bTrain: true);
            ulong[] testAllBits = AllFeatures.ReadAllBits(bTrain: false);

            var resLabs = CalcPartialBayes(labels, trainAllBits, testAllBits,
                new int[] { 0, 2, 31, 5, 37, 35, 14, 30, 32, 7, 12, 9, 35, 6, 39 + 1, 39 + 2, 39 + 3, 39 + 4, 39 + 5, 17, 33, 8 },
                bOne: false);

            return resLabs;
        }
    }
}
