﻿using System;
using System.Collections.Generic;
using System.Linq;
using GenEngAttr.Data;
using GenEngAttr.Features;
using GenEngAttr.Utils;

namespace GenEngAttr.Models
{
    // plasmid to plasmid similarity matrixes
    public static class PlasmidsSim
    {
        // same as T3Kernel, but without float[] weights
        public static ushort[][] PlasmidsSimKernel(int trainLength, int testLength,
            int[] trainPolymers, ushort[] trainPlasmids, int[] testPolymers, ushort[] testPlasmids, int[][] devSetTops)
        {
            ushort[][] counts = Util.InitArray(0, testLength).Select(v => new ushort[trainLength]).ToArray();

            int iStart = 0;
            int jStart = 0;

            for (int i = 0; i < testPolymers.Length; i++)
            {
                if ((i & 0xFFFFF) == 0)
                {
                    Util.Print("PlasmidsSimKernel {0}/{1} {2}", i, testPolymers.Length, DateTime.Now.MyToSecString());
                }

                if ((i == testPolymers.Length - 1) || (testPolymers[i] != testPolymers[i + 1]))
                {
                    while ((jStart < trainPolymers.Length) && (trainPolymers[jStart] < testPolymers[i])) jStart++;

                    if (jStart == trainPolymers.Length)
                    {
                        break;
                    }

                    if (trainPolymers[jStart] != testPolymers[i])
                    {
                        iStart = i + 1;
                        continue;
                    }

                    int jEnd = jStart;

                    while ((jEnd < trainPolymers.Length) && (trainPolymers[jEnd] == testPolymers[i])) jEnd++;

                    for (int ik = iStart; ik <= i; ik++)
                    {
                        bool bNewPolymer = true;

                        for (int l = 0; bNewPolymer && (l < devSetTops.Length); l++)
                        {
                            for (int j = jStart; j < jEnd; j++)
                            {
                                if (trainPlasmids[j] == devSetTops[l][testPlasmids[ik]])
                                {
                                    bNewPolymer = false;
                                    break;
                                }
                            }
                        }

                        if (bNewPolymer)
                        {
                            for (int j = jStart; j < jEnd; j++)
                            {
                                counts[testPlasmids[ik]][trainPlasmids[j]] += 1;
                            }
                        }
                    }

                    iStart = i + 1;
                    jStart = jEnd;
                }
            }

            return counts;
        }

        public static ushort[][] GetPlasmidsSimHybMT(int nGramLength, int[][] resPlasmidTops)
        {
            bool[] disables = Util.InitArray(true, DataSet.NTest);

            var (trainPolymers, trainPlasmids) = AllFeatures.ReadTransform2(bTrain: true, nGramLength: nGramLength);

            var (polymers, plasmids) = AllFeatures.ReadTransform2(bTrain: false, nGramLength: nGramLength);

            ushort[][][] res = new ushort[12][][];

            Util.MyParallelFor(DataSet.NThreads6, 0, res.Length, (i) =>
            {
                var (partDisables, nDisable) = GAUtil.GetPartDisables(disables, i, res.Length);

                var (devt2Polymers, devt2Plasmids) = Transform2.Transform2GetOnlyDisables(polymers, plasmids, partDisables);

                int[][] devSetTops = GAUtil.GetTops(resPlasmidTops, partDisables);

                res[i] = PlasmidsSimKernel(DataSet.NTrain, nDisable, trainPolymers, trainPlasmids, devt2Polymers, devt2Plasmids, devSetTops);
            });

            List<ushort[]> resLs = new List<ushort[]>();

            for (int i = 0; i < res.Length; i++)
            {
                resLs.AddRange(res[i]);
            }

            return resLs.ToArray();
        }

        public static int[] GetMaxSimPlasmids(int nGramLength)
        {
            double mul = 1e-8f;
            ushort[][] resLabs = AllModels.ReadSimPlasmids(nGramLength: nGramLength);

            int[] labels = DataSet.ReadLabels();
            int[] labsCounts = DataSet.ReadLabelsCountsNotSafeForCV();
            double[] trainCounts = new double[DataSet.NTrain];

            for (int i = 0; i < trainCounts.Length; i++)
            {
                trainCounts[i] = labsCounts[labels[i]] * mul;
            }

            int[] results = new int[resLabs.Length];

            for (int i = 0; i < resLabs.Length; i++) 
            {
                double maxValue = resLabs[i][0] + trainCounts[0];
                results[i] = 0;

                for (int j = 1; j < resLabs[i].Length; j++)
                {
                    double value = resLabs[i][j] + trainCounts[j];

                    if (maxValue < value)
                    {
                        maxValue = value;
                        results[i] = j;
                    }
                }
            }

            return results;
        }

        public static ushort[][] GetPlasmidsSim(int nGramLength)
        {
            return GetPlasmidsSimHybMT(nGramLength, new int[][] { });
        }

        public static ushort[][] GetPlasmidsSimHyb1(int nGramLength)
        {
            var resLabsTop1 = GetMaxSimPlasmids(nGramLength);

            return GetPlasmidsSimHybMT(nGramLength, new int[][] { resLabsTop1 });
        }
    }
}
