﻿using System;
using System.Collections.Generic;
using System.Linq;
using GenEngAttr.Data;
using GenEngAttr.Utils;

namespace GenEngAttr.Models
{
    // soft mask model
    public static class SoftMasks
    {
        public static void ApplySoftMasks(double[][] resLabs, double pow = 0.5)
        {
            int[] labels = DataSet.ReadLabels();
            ulong[] trainColumnsULong = DataSet.ReadColumnsULong(bTrain: true);
            ulong[] testColumnsULong = DataSet.ReadColumnsULong(bTrain: false);
            Dictionary<ulong, Dictionary<int, int>> bitDictionary = new Dictionary<ulong, Dictionary<int, int>>();
            Dictionary<ulong, Tuple<int, int>> bitDictionaryCounts = new Dictionary<ulong, Tuple<int, int>>();
            Dictionary<ulong, KeyValuePair<int, int>[]> forceLabsDictionary = new Dictionary<ulong, KeyValuePair<int, int>[]>();

            for (int i = 0; i < trainColumnsULong.Length; i++)
            {
                if (!bitDictionary.ContainsKey(trainColumnsULong[i]))
                {
                    bitDictionary[trainColumnsULong[i]] = new Dictionary<int, int>();
                }

                Dictionary<int, int> labDictionary = bitDictionary[trainColumnsULong[i]];

                if (labDictionary.ContainsKey(labels[i]))
                {
                    labDictionary[labels[i]]++;
                }
                else
                {
                    labDictionary[labels[i]] = 1;
                }
            }

            foreach (var keyValue in bitDictionary)
            {
                ulong bits = keyValue.Key;
                Dictionary<int, int> labDictionary = keyValue.Value;

                int labCount = labDictionary.Count();
                int totalCount = labDictionary.Select(v => v.Value).Sum();

                bitDictionaryCounts.Add(bits, Util.MyTuple(labCount, totalCount));

                if ((totalCount > 0))
                {
                    int lab = labDictionary.First().Key;

                    var labs = labDictionary.OrderByDescending(v => v.Value).ToArray();

                    forceLabsDictionary[bits] = labs;
                }
            }

            for (int i = 0; i < resLabs.Length; i++)
            {
                ulong bits = testColumnsULong[i];

                if (!forceLabsDictionary.ContainsKey(bits))
                {
                    List<KeyValuePair<int, int>> labsList = new List<KeyValuePair<int, int>>();

                    for (int j = 2; j < 20; j++)
                    {
                        foreach (var keyValue in forceLabsDictionary)
                        {
                            ulong newBits = keyValue.Key;

                            newBits = newBits ^ bits;

                            if (Util.PopCount(newBits) < j)
                            {
                                labsList.AddRange(keyValue.Value);
                            }
                        }

                        if (labsList.Count > 0)
                        {
                            break;
                        }
                    }

                    forceLabsDictionary[bits] = labsList.OrderByDescending(v => v.Value).ToArray();
                }

                var labs = forceLabsDictionary[bits];
                double alpha = Math.Pow(labs.Length, pow);

                for (int j = 0; j < labs.Length; j++)
                {
                    resLabs[i][labs[j].Key] += 1d / (alpha + j);
                }
            }
        }
    }
}
