﻿using System;
using System.Collections.Generic;
using GenEngAttr.Data;
using GenEngAttr.Features;
using GenEngAttr.Utils;

namespace GenEngAttr.Models
{
    class T3Models
    {
        public static double[][] T3Kernel(int trainLength, int testLength,
                int[] trainPolymers, int[] trainLabs, float[] weights, int[] testPolymers, ushort[] testPlasmids, int[][] devSetTops)
        {
            double[][] resLabs = new double[testLength][];

            for (int i = 0; i < resLabs.Length; i++)
            {
                resLabs[i] = new double[trainLength];
            }

            int iStart = 0;
            int jStart = 0;

            for (int i = 0; i < testPolymers.Length; i++)
            {
                if ((i == testPolymers.Length - 1) || (testPolymers[i] != testPolymers[i + 1]))
                {
                    while ((jStart < trainPolymers.Length) && (trainPolymers[jStart] < testPolymers[i])) jStart++;

                    if (jStart == trainPolymers.Length)
                    {
                        break;
                    }

                    if (trainPolymers[jStart] != testPolymers[i])
                    {
                        iStart = i + 1;
                        continue;
                    }

                    int jEnd = jStart;

                    while ((jEnd < trainPolymers.Length) && (trainPolymers[jEnd] == testPolymers[i])) jEnd++;

                    for (int ik = iStart; ik <= i; ik++)
                    {
                        bool bNewPolymer = true;

                        for (int l = 0; bNewPolymer && (l < devSetTops.Length); l++)
                        {
                            for (int j = jStart; j < jEnd; j++)
                            {
                                if (trainLabs[j] == devSetTops[l][testPlasmids[ik]])
                                {
                                    bNewPolymer = false;
                                    break;
                                }
                            }
                        }

                        if (bNewPolymer)
                        {
                            for (int j = jStart; j < jEnd; j++)
                            {
                                resLabs[testPlasmids[ik]][trainLabs[j]] += weights[j];
                            }
                        }
                    }

                    iStart = i + 1;
                    jStart = jEnd;
                }

                if ((i & 0xFFFFF) == 0)
                {
                    //Util.Print("T3Kernel {0}/{1} {2}", i, testPolymers.Length, DateTime.Now.MyToLongString());
                }
            }

            return resLabs;
        }

        public static double[][] GetT3HybMT(int nGramLength, int[][] resLabsTops)
        {
            bool[] disables = Util.InitArray(true, DataSet.NTest);

            var (trainPolymers, trainLabs, trainCounts, trainShares) = AllFeatures.ReadTransform3(nGramLength: nGramLength);

            var (polymers, plasmids) = AllFeatures.ReadTransform2(bTrain: false, nGramLength: nGramLength);

            double[][][] res = new double[12][][];

            Util.MyParallelFor(4, 0, res.Length, (i) =>
            {
                var (partDisables, nDisable) = GAUtil.GetPartDisables(disables, i, res.Length);

                var (devt2Polymers, devt2Plasmids) = Transform2.Transform2GetOnlyDisables(polymers, plasmids, partDisables);

                int[][] devSetTops = GAUtil.GetTops(resLabsTops, partDisables);

                res[i] = T3Kernel(DataSet.NLabs, nDisable, trainPolymers, trainLabs, trainShares, devt2Polymers, devt2Plasmids, devSetTops);
            });

            List<double[]> resLs = new List<double[]>();

            for (int i = 0; i < res.Length; i++)
            {
                resLs.AddRange(res[i]);
            }

            return resLs.ToArray();
        }

        public static double[][] GetT3(int nGramLength)
        {
            return GetT3HybMT(nGramLength: nGramLength, resLabsTops: new int[][] { });
        }

        public static double[][] GetT3Hyb1(int nGramLength)
        {
            var resLabs = AllModels.ReadT3(nGramLength);

            int[] resLabsTop1 = GAUtil.GetResLabsTop1(resLabs);

            return GetT3HybMT(nGramLength: nGramLength, resLabsTops: new int[][] { resLabsTop1 });
        }
    }
}
