﻿using GenEngAttr.Features;
using GenEngAttr.Utils;
using System;
using System.Linq;

namespace GenEngAttr.Models
{
    // waves are several models in pipeline
    public static class WavesModels
    {
        public static double[][] GetWaveA(int plGramLength)
        {
            int t3NGramLength = 19;

            var resLabsT3 = AllModels.ReadT3(t3NGramLength);

            var resLabsT3Hyb1 = AllModels.ReadT3Hyb1(t3NGramLength);

            var resLabsPl = AllModels.ReadPlasmids(plGramLength);

            var resLabsPlHyb1 = AllModels.ReadPlasmidsHyb1(plGramLength);

            var resLabs = Merges.PolyMerge2(resLabsT3, resLabsT3Hyb1, resLabsPl, resLabsPlHyb1,
                AllFeatures.ReadPersonalPlasmid(bTrain: false, nGramLength: plGramLength, iPlasmid: 0),
                0.64, 0.63, 0.53, 0.76);

            var resLabsBayes16 = AllModels.ReadWeightBayes16();

            resLabs = Merges.BayesMerge1(resLabs, resLabsBayes16, resLabsPl, 33, 1);

            double pow = 0.24;

            double[][] partialBayes = AllModels.ReadPartialBayes();

            for (int i = 0; i < partialBayes.Length; i++)
            {
                double sum = partialBayes[i].Sum();

                for (int j = 0; j < partialBayes[i].Length; j++)
                {
                    partialBayes[i][j] = resLabs[i][j] * Math.Pow(partialBayes[i][j] / sum, pow);
                }
            }

            resLabs = Merges.CombMerg1(partialBayes, resLabs, resLabsPl, 0.39, 7100);

            return resLabs;
        }

        public static double[][] GetWaveB(int plGramLength) 
        {
            var resLabs = AllModels.ReadWaveA(plGramLength);

            AllModels.ApplySoftMasks(resLabs);

            return resLabs;
        }

        public static int[][] GetWaveCTop10(int plGramLength) 
        {
            var resLabs = AllModels.ReadWaveB(plGramLength);

            resLabs = Merges.ResLabsToRevRank(resLabs);

            resLabs = HighPlasmidsClusters.ApplyHighPlasmidsClusters(resLabs, plGramLength, 1.98);

            resLabs = Merges.ResLabsToRevRank(resLabs);

            for (int i = 0; i < resLabs.Length; i++)
            {
                resLabs[i][666] += 0.12f;
            }

            int[][] resTop10 = GAUtil.GetResTop10(resLabs);

            return resTop10;
        }

        public static int[][] GetWaveDTop10(int plGramLength) 
        {
            var resLabs = AllModels.ReadWaveA(plGramLength);

            int[][] top10 = AllModels.ReadWaveCTop10(plGramLength);

            AllModels.ApplyHighSoftMasks(resLabs, top10);

            resLabs = Merges.ResLabsToRevRank(resLabs);

            resLabs = HighPlasmidsClusters.ApplyHighPlasmidsClusters(resLabs, plGramLength, 1.98);

            resLabs = Merges.ResLabsToRevRank(resLabs);

            for (int i = 0; i < resLabs.Length; i++)
            {
                resLabs[i][666] += 0.12f;
            }

            int[][] resTop10 = GAUtil.GetResTop10(resLabs);

            return resTop10;
        }
    }
}
