﻿using System;
using System.Collections.Generic;
using System.Linq;
using GenEngAttr.Data;
using GenEngAttr.Features;
using GenEngAttr.Utils;

namespace GenEngAttr.Models
{
    class WeightBayes
    {
        private static double[] GetWeights(int[] labels, int[] kds, ulong[] trainBits,
            double[] totalLabsCount, double[][] oneVals, double[][] zeroVals, bool bOne, int trainPartLength, int nTries)
        {
            List<double> dst = new List<double>();
            List<double> curSrc = new List<double>();
            List<double>[] bitSrc = Util.InitArray(0, kds.Length).Select(v => new List<double>()).ToArray();
            double[] weights = new double[kds.Length + 1];
            double[][] logOneVals = Util.InitArray(0, kds.Length).Select(v => new double[DataSet.NLabs]).ToArray();
            double[][] logZeroVals = Util.InitArray(0, kds.Length).Select(v => new double[DataSet.NLabs]).ToArray();
            double[][] trainVals = null;// MyMath.InitArray(0, DataSet.NTrain).Select(v => new double[DataSet.NLabs]).ToArray();

            {
                double[] resLabs0 = new double[DataSet.NLabs];

                for (int j = 0; j < resLabs0.Length; j++)
                {
                    resLabs0[j] = bOne ? 1d : Math.Max(1, totalLabsCount[j]);
                }

                double total = Math.Max(1, resLabs0.Sum());

                for (int j = 0; j < resLabs0.Length; j++)
                {
                    resLabs0[j] = Math.Log(resLabs0[j] / total);
                }

                trainVals = Util.InitArray(0, DataSet.NTrain).Select(v => resLabs0.ToArray()).ToArray();
            }

            for (int i = 0; i < kds.Length; i++)
            {
                for (int j = 0; j < logOneVals[i].Length; j++)
                {
                    logOneVals[i][j] = Math.Log(oneVals[i][j]);
                    logZeroVals[i][j] = Math.Log(zeroVals[i][j]);
                }
            }

            for (int iTry = 0; iTry < nTries; iTry++)
            {
                for (int k = 0; k < kds.Length; k++)
                {
                    int n = trainPartLength * DataSet.NLabs;
                    double a = 0;
                    double b = 0; // a + b * x
                    double sumx = 0;  // sum of x
                    double sumx2 = 0; // sum of x**2
                    double sumxy = 0; // sum of x * y
                    double sumy = 0;  // sum of y               
                    double prec = -DataSet.NLabs + 1; // why not xD

                    for (int i = 0; i < DataSet.NTrain; i++)
                    {
                        double[] mulVals = ((1 == (int)((trainBits[i] >> kds[k]) & (1UL))) ? logOneVals : logZeroVals)[k];
                        int thisLab = labels[i];

                        for (int j = 0; j < DataSet.NLabs; j++)
                        {
                            double yI = ((j == thisLab) ? 0 : prec) - trainVals[i][j];

                            sumx += mulVals[j];
                            sumx2 += mulVals[j] * mulVals[j];
                            sumxy += mulVals[j] * yI;
                            sumy += yI;
                        }
                    }

                    double denom = (n * sumx2 - sumx * sumx);

                    if (denom != 0)
                    {
                        b = (n * sumxy - sumx * sumy) / denom;
                        a = (sumy * sumx2 - sumx * sumxy) / denom;
                    }
                    else
                    {
                        a = sumy / n;
                    }

                    //if (b > 0)
                    {
                        b = Math.Min(1d / nTries, b);
                        b = Math.Max(0 - weights[k + 1], b);

                        weights[0] += a;
                        weights[k + 1] += b;

                        for (int i = 0; i < DataSet.NTrain; i++)
                        {
                            double[] mulVals = ((1 == (int)((trainBits[i] >> kds[k]) & (1UL))) ? logOneVals : logZeroVals)[k];
                            int thisLab = labels[i];

                            for (int j = 0; j < DataSet.NLabs; j++)
                            {
                                trainVals[i][j] += a + b * mulVals[j];
                            }
                        }
                    }

                    //Util.Print("Submit Bayes Weights k: {0} a: {1} b: {2} total weight: {3}", k, a, b, weights[k + 1]);
                }
            }

            return weights;
        }

        public static double[][] CalcWeightBayes(double[][] resLabs, int[] labels, ulong[] trainBits, ulong[] testBits, int[] kds, bool bOne = true, int nTries = 16)
        {
            double alphaRegulator = 1;
            double[] totalLabsCount = new double[DataSet.NLabs];
            double[][] labsCounters = Util.InitArray(0, kds.Length).Select(v => new double[DataSet.NLabs]).ToArray();
            double[][] oneVals = Util.InitArray(0, kds.Length).Select(v => new double[DataSet.NLabs]).ToArray();
            double[][] zeroVals = Util.InitArray(0, kds.Length).Select(v => new double[DataSet.NLabs]).ToArray();
            double[] columnsCounters = new double[kds.Length];
            double totalCount = 0;

            for (int i = 0; i < trainBits.Length; i++)
            {
                int thisLab = labels[i];

                for (int k = 0; k < kds.Length; k++)
                {
                    int value = (int)((trainBits[i] >> kds[k]) & (1UL));

                    labsCounters[k][thisLab] += value;
                    columnsCounters[k] += value;
                }

                totalLabsCount[thisLab]++;
                totalCount++;
            }

            for (int j = 0; j < DataSet.NLabs; j++)
            {
                for (int k = 0; k < kds.Length; k++)
                {
                    labsCounters[k][j] += alphaRegulator;
                    columnsCounters[k] += alphaRegulator;
                }
                totalLabsCount[j] += 2 * alphaRegulator;
                totalCount += 2 * alphaRegulator;
            }

            for (int j = 0; j < DataSet.NLabs; j++)
            {
                double divider = (1d / Math.Max(1, totalLabsCount[j]));

                for (int k = 0; k < kds.Length; k++)
                {
                    oneVals[k][j] = (labsCounters[k][j]) / (columnsCounters[k]) * divider;
                    zeroVals[k][j] = ((totalLabsCount[j] - labsCounters[k][j])) / ((totalCount - columnsCounters[k])) * divider;
                }
            }

            if (resLabs == null)
            {
                resLabs = new double[testBits.Length][];

                double[] resLabs0 = new double[DataSet.NLabs];

                for (int j = 0; j < resLabs0.Length; j++)
                {
                    resLabs0[j] = bOne ? 1d : Math.Max(1, totalLabsCount[j]);
                }

                for (int i = 0; i < testBits.Length; i++)
                {
                    resLabs[i] = resLabs0.ToArray();
                }
            }

            double[] weights = GetWeights(labels, kds, trainBits, totalLabsCount, oneVals, zeroVals, bOne, trainBits.Length, nTries);

            for (int i = 0; i < testBits.Length; i++)
            {
                double[] resLabsI = resLabs[i];

                for (int k = 0; k < kds.Length; k++)
                {
                    double[] mulVals = ((((testBits[i] >> kds[k]) & (1UL)) == 1) ? oneVals : zeroVals)[k];

                    for (int j = 0; j < resLabsI.Length; j++)
                    {
                        resLabsI[j] *= Math.Pow(mulVals[j], weights[k + 1]);
                    }
                }
            }

            return resLabs;
        }

        public static double[][] GetWeightBayes16()
        {
            int[] labels = DataSet.ReadLabels();
            ulong[] trainAllBits = AllFeatures.ReadAllBits(bTrain: true);
            ulong[] testAllBits = AllFeatures.ReadAllBits(bTrain: false);

            var weightBayes16 = CalcWeightBayes(null, labels, trainAllBits, testAllBits,
                Util.InitArrayStep(0, 1, DataSet.NColumnsAll), bOne: false, nTries: 16);

            GAUtil.NormalizeResLabs(weightBayes16);

            return weightBayes16;
        }
    }
}
