﻿using System;
using System.Diagnostics;
using System.Globalization;
using System.Linq;
using System.Threading;
using GenEngAttr.Data;
using GenEngAttr.Features;
using GenEngAttr.Models;
using GenEngAttr.Utils;

namespace GenEngAttr
{
    // main point, creating submit file, Submit201019() is generating final model in 40 min
    class Program
    {
        public static void PrintUsage()
        {
            Util.Print(@"Usage: GenEngAttr.exe c:\GenEngAttr\data");
            Util.Print(@"       GenEngAttr.exe -innovation1 c:\GenEngAttr\data");
            Util.Print(@"       GenEngAttr.exe -innovation2 c:\GenEngAttr\data");
            Util.Print(@"       GenEngAttr.exe -innovation3 c:\GenEngAttr\data");
            Util.Print(@"       GenEngAttr.exe -innovation4 c:\GenEngAttr\data");
            Util.Print("");
            Util.Print("Tested on Windows-7/MVSC-2017/.Net-4.8/x64  16Gb/i7-3770K/SSD");
            Util.Print("Expect creation of ~20Gb in \"data\\cache\" directory");
            Util.Print("Clean the \"data\\cache\" directory if change the test file or want a fresh start");
            Util.Print("Expect 40 min for a full run down to 4 min for -innovation1 run");
            Util.Print("Times may vary based on runtime optimizations, memory timings, SSD speed, etc");
            Util.Print("The resulting file will be saved to \"data\\submits\" directory");
        }

        static void Main(string[] args)
        {
            DateTime now = DateTime.Now;

            Process.GetCurrentProcess().PriorityClass = ProcessPriorityClass.BelowNormal;
            CultureInfo.DefaultThreadCurrentCulture = CultureInfo.InvariantCulture;

            if (DataSet.SetupDataDirectory(args))
            {
                if ((args.Length > 0) && (args[0] == "-innovation1"))
                {
                    InnovationTrack1();
                }
                else
                if ((args.Length > 0) && (args[0] == "-innovation2"))
                {
                    InnovationTrack2();
                }
                else
                if ((args.Length > 0) && (args[0] == "-innovation3"))
                {
                    InnovationTrack3();
                }
                else
                if ((args.Length > 0) && (args[0] == "-innovation4"))
                {
                    InnovationTrack4();
                }
                else
                {
                    Submit201019(); 
                }

                Util.Print("Total Time: {0}", DateTime.Now - now);
            }
            else
            {
                PrintUsage();
            }

            Util.Print("Press any key to exit...");

            Console.ReadKey();
        }

        public static void InnovationTrack1()
        {
            Util.Print("Expect ~4m08s min working time from fresh start on Intel i7-3770K + SSD"); 

            int t3GramLength = 19;

            var resT3 = AllModels.ReadT3(t3GramLength: t3GramLength);

            DataSet.SaveResults(resT3, bSave: true);
        }

        public static void InnovationTrack2()
        {
            Util.Print("Expect ~4m40s min working time from fresh start on Intel i7-3770K + SSD"); 

            int t3GramLength = 19;

            var resT3 = AllModels.ReadT3(t3GramLength: t3GramLength);

            var resT3Hyb1 = AllModels.ReadT3Hyb1(t3GramLength: t3GramLength);

            var resLabs = Merges.PolyMerge4(resT3, resT3Hyb1, 0.62);

            DataSet.SaveResults(resLabs, bSave: true);
        }

        public static void InnovationTrack3()
        {
            Util.Print("Expect ~26m18s working time from fresh start on Intel i7-3770K + SSD"); 

            int plGramLength = 15;
            int t3GramLength = 19;

            var resT3 = AllModels.ReadT3(t3GramLength: t3GramLength);

            var resLabsT3Hyb1 = AllModels.ReadT3Hyb1(t3GramLength: t3GramLength);

            var resLabsPl = AllModels.ReadPlasmids(plGramLength: plGramLength);

            var resLabsPlHyb1 = AllModels.ReadPlasmidsHyb1(plGramLength: plGramLength);

            var resLabs = Merges.PolyMerge2(resT3, resLabsT3Hyb1, resLabsPl, resLabsPlHyb1,
                AllFeatures.ReadPersonalPlasmid(bTrain: false, nGramLength: plGramLength, iPlasmid: 0),
                0.64, 0.63, 0.53, 0.76);

            resLabs = AllModels.ApplyHighPlasmidsClusters(resLabs, plGramLength, 1.98);

            DataSet.SaveResults(resLabs, bSave: true);
        }

        public static void InnovationTrack4()
        {
            Util.Print("Expect ~4m58s working time from fresh start on Intel i7-3770K + SSD"); 

            int t3GramLength = 19;

            var resT3 = AllModels.ReadT3(t3GramLength: t3GramLength);

            var resT3Hyb1 = AllModels.ReadT3Hyb1(t3GramLength: t3GramLength);

            var resLabs = Merges.PolyMerge4(resT3, resT3Hyb1, 0.62);

            var partialBayes = AllModels.ReadPartialBayes();

            double pow = 0.24;

            for (int i = 0; i < resLabs.Length; i++)
            {
                double sum = partialBayes[i].Sum();

                for (int j = 0; j < partialBayes[i].Length; j++)
                {
                    partialBayes[i][j] = resLabs[i][j] * Math.Pow(partialBayes[i][j] / sum, pow);
                }
            }

            resLabs = Merges.RankSum2(partialBayes, resLabs, 0.9);

            AllModels.ApplySoftMasks(resLabs);

            DataSet.SaveResults(resLabs, bSave: true);
        }

        public static void Submit201019()
        {
            Util.Print("Expect ~40m26s working time from fresh start on Intel i7-3770K + SSD"); 

            int plGramLength = 15;
            int t3GramLength = 19;

            var resHighT3 = AllModels.ReadHighT3(t3GramLength: t3GramLength);

            var resLabsT3Hyb1 = AllModels.ReadT3Hyb1(t3GramLength: t3GramLength);

            var resLabsPl = AllModels.ReadPlasmids(plGramLength: plGramLength);

            var resLabsPlHyb1 = AllModels.ReadPlasmidsHyb1(plGramLength: plGramLength);

            var resLabs = Merges.PolyMerge3(resHighT3, resLabsT3Hyb1, resLabsPl, resLabsPlHyb1,
                AllFeatures.ReadPersonalPlasmid(bTrain: false, nGramLength: plGramLength, iPlasmid: 0),
                0.64, 0.62, 0.56, 0.76);

            var resLabsBayes16 = AllModels.ReadWeightBayes16();

            resLabs = Merges.BayesMerge1(resLabs, resLabsBayes16, resLabsPl, 10, 1);

            var highPartialBayes = AllModels.ReadHighPartialBayes();

            double pow = 0.24;

            for (int i = 0; i < resLabs.Length; i++)
            {
                double sum = highPartialBayes[i].Sum();

                for (int j = 0; j < highPartialBayes[i].Length; j++)
                {
                    highPartialBayes[i][j] = resLabs[i][j] * Math.Pow(highPartialBayes[i][j] / sum, pow);
                }
            }

            resLabs = Merges.CombMerg1(highPartialBayes, resLabs, resLabsPl, 0.9, 9500);

            int[][] top10 = AllModels.ReadWaveCTop10(plGramLength);

            AllModels.ApplyHighSoftMasks(resLabs, top10, beta: 0.04);

            resLabs = Merges.ResLabsToRevRank(resLabs);

            resLabs = AllModels.ApplyHighPlasmidsClusters(resLabs, plGramLength, 2.47);

            resLabs = Merges.ResLabsToRevRank(resLabs);

            for (int i = 0; i < resLabs.Length; i++)
            {
                resLabs[i][666] += 0.12f;
            }

            DataSet.SaveResults(resLabs, bSave: true);
        }
    }
}
