﻿using System;
using System.Linq;
using GenEngAttr.Data;
using GenEngAttr.Features;

namespace GenEngAttr.Utils
{
    // Genetic Engineering Attribution specific utility methods
    public static class GAUtil
    {
        public static int GetNum(char v)
        {
            switch (v)
            {
                case 'A': return 0;
                case 'C': return 1;
                case 'G': return 2;
                case 'T': return 3;
                case 'N': return 4;
                default:
                    return 26;
            }
        }

        public static char GetChar(int v)
        {
            char[] table = new char[] { 'A', 'C', 'G', 'T', 'N' };

            return table[v];
        }

        public static string GetString(int v, int nGramLength)
        {
            string res = "";

            for (int i = 0; i < nGramLength; i++)
            {
                res = GetChar(v & 3) + res;
                v >>= 2;
            }

            return res;
        }

        public static Tuple<bool[], int> GetPartDisables(bool[] disables, int iPart, int nParts)
        {
            bool[] res = new bool[disables.Length];
            int nPiece = (res.Length + nParts - 1) / nParts;
            int iStart = iPart * nPiece;
            int iEnd = Math.Min(res.Length, iStart + nPiece);
            int count = 0;

            for (int i = iStart; i < iEnd; i++)
            {
                res[i] = disables[i];
                count += (disables[i] ? 1 : 0);
            }

            return Util.MyTuple(res, count);
        }

        public static int[] GetResLabsTop1(double[][] resLabs)
        {
            int[] results = new int[resLabs.Length];

            for (int i = 0; i < resLabs.Length; i++)
            {
                double maxValue = resLabs[i][0];
                results[i] = 0;

                for (int j = 1; j < resLabs[i].Length; j++)
                {
                    if (maxValue < resLabs[i][j])
                    {
                        maxValue = resLabs[i][j];
                        results[i] = j;
                    }
                }
            }

            return results;
        }

        public static int[][] GetDevSetTops(int[][] resLabsTops, int[] devSet, bool[] partDisables)
        {
            int n = 0;

            for (int i = 0; i < partDisables.Length; i++)
            {
                if (partDisables[i])
                {
                    n++;
                }
            }

            int[][] devLabsTops = Util.InitArray(0, resLabsTops.Length).Select(v => new int[n]).ToArray();
            int offset = 0;

            for (int id = 0; id < devSet.Length; id++)
            {
                if (partDisables[devSet[id]])
                {
                    for (int i = 0; i < resLabsTops.Length; i++)
                    {
                        devLabsTops[i][offset] = resLabsTops[i][id];
                    }

                    offset++;
                }
            }

            return devLabsTops;
        }

        public static int[][] GetTestSetTops(int[][] resLabsTops, bool[] partDisables)
        {
            int n = 0;

            for (int i = 0; i < partDisables.Length; i++)
            {
                if (partDisables[i])
                {
                    n++;
                }
            }

            int[][] testLabsTops = Util.InitArray(0, resLabsTops.Length).Select(v => new int[n]).ToArray();
            int offset = 0;

            for (int id = 0; id < partDisables.Length; id++)
            {
                if (partDisables[id])
                {
                    for (int i = 0; i < resLabsTops.Length; i++)
                    {
                        testLabsTops[i][offset] = resLabsTops[i][id];
                    }

                    offset++;
                }
            }

            return testLabsTops;
        }

        public static int[][] GetTops(int[][] resLabsTops, bool[] partDisables)
        {
            int n = 0;

            for (int i = 0; i < partDisables.Length; i++)
            {
                if (partDisables[i])
                {
                    n++;
                }
            }

            int[][] devLabsTops = Util.InitArray(0, resLabsTops.Length).Select(v => new int[n]).ToArray();
            int offset = 0;

            for (int id = 0; id < partDisables.Length; id++)
            {
                if (partDisables[id])
                {
                    for (int i = 0; i < resLabsTops.Length; i++)
                    {
                        devLabsTops[i][offset] = resLabsTops[i][id];
                    }

                    offset++;
                }
            }

            return devLabsTops;
        }

        public static void NormalizeResLabs(double[][] resLabsCombProp)
        {
            for (int i = 0; i < resLabsCombProp.Length; i++)
            {
                double revSum = 1d / resLabsCombProp[i].Sum();

                if ((Double.IsNaN(revSum) || Double.IsInfinity(revSum)))
                {
                    Util.Print("!!! NormalizeResLabs Fail {0} {1}", i, revSum);
                }

                for (int j = 0; j < resLabsCombProp[i].Length; j++)
                {
                    resLabsCombProp[i][j] = resLabsCombProp[i][j] * revSum;
                }
            }
        }

        public static int[] AppendDevSet(int[] devSet)
        {
            int[] newDevSet = new int[devSet.Length + DataSet.NTest];

            devSet.CopyTo(newDevSet, 0);

            for (int i = 0; i < DataSet.NTest; i++)
            {
                newDevSet[i + devSet.Length] = i + DataSet.NTrain;
            }

            return newDevSet;
        }

        public static int[] AppendLabels(int[] labels)
        {
            int[] newLabels = new int[labels.Length + DataSet.NTest];

            labels.CopyTo(newLabels, 0);

            for (int i = 0; i < DataSet.NTest; i++)
            {
                newLabels[i + DataSet.NTrain] = -1;
            }

            return newLabels;
        }

        public static Tuple<int[], int[]> GetOutHighLabelsAndDevSet()
        {
            int[] labels = DataSet.ReadLabels();
            int[] devSet = AppendDevSet(new int[] { });

            labels = AppendLabels(labels);

            return Util.MyTuple(labels, devSet);
        }

        public static int[][] GetResTop10(double[][] resLabs)
        {
            int topN = DataSet.TopN;

            int[][] resTop10 = Util.InitArray(0, topN).Select(v => new int[resLabs.Length]).ToArray();

            for (int i = 0; i < resLabs.Length; i++)
            {
                int[] orders = resLabs[i].MyOrderDescending();

                for (int j = 0; j < topN; j++)
                {
                    resTop10[j][i] = orders[j];
                }
            }

            return resTop10;
        }

        public static ulong[] GetHighAllBits()
        {
            ulong[] trainAllBits = AllFeatures.ReadAllBits(bTrain: true);
            ulong[] testAllBits = AllFeatures.ReadAllBits(bTrain: false);

            ulong[] highBitsAll = new ulong[trainAllBits.Length + testAllBits.Length];

            trainAllBits.CopyTo(highBitsAll, 0);
            testAllBits.CopyTo(highBitsAll, trainAllBits.Length);

            return highBitsAll;
        }

    }
}
