﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace GenEngAttr.Utils
{
    // various models' merges
    class Merges
    {
        public static double[][] ResLabsToRevRank(double[][] src)
        {
            double[][] dst = new double[src.Length][];

            for (int i = 0; i < src.Length; i++)
            {
                int[] orders1 = src[i].MyOrderDescending();

                dst[i] = new double[src[i].Length];

                for (int j = 0; j < orders1.Length; j++)
                {
                    dst[i][orders1[j]] += 1d / (j + 1d);
                }
            }

            return dst;
        }

        public static double[][] RankSum2(double[][] src1, double[][] src2, double alpha)
        {
            double[][] dst = new double[src1.Length][];

            for (int i = 0; i < src1.Length; i++)
            {
                int[] orders1 = src1[i].MyOrderDescending();
                int[] orders2 = src2[i].MyOrderDescending();

                dst[i] = new double[src1[i].Length];

                for (int j = 0; j < orders1.Length; j++)
                {
                    dst[i][orders1[j]] += alpha / (j + 1f);
                    dst[i][orders2[j]] += (1 - alpha) / (j + 1f);
                }
            }

            return dst;
        }

        public static double[][] CombMerg1(double[][] combLabs, double[][] resLabs, double[][] Pl, double alpha, double divider)
        {
            double[][] dst = new double[Pl.Length][];

            for (int i = 0; i < Pl.Length; i++)
            {
                int[] orders1 = combLabs[i].MyOrderDescending();
                double plMax = Pl[i].Max();
                double thisAlpha = 1 - alpha * Math.Min(1, plMax / divider);

                dst[i] = new double[Pl[i].Length];

                for (int j = 0; j < dst[i].Length; j++)
                {
                    dst[i][orders1[j]] += thisAlpha / (j + 1f);
                    dst[i][j] += (1 - thisAlpha) * resLabs[i][j];
                }
            }

            return dst;
        }

        public static double[][] BayesMerge1(double[][] poly, double[][] bayes, double[][] Pl, double alpha1, double alpha2)
        {
            double[][] dst = new double[Pl.Length][];

            for (int i = 0; i < Pl.Length; i++)
            {
                int[] orders2 = bayes[i].MyOrderDescending();

                double plMax = Pl[i].Max();
                double thisAlpha = Math.Min((plMax / alpha1), alpha2); // (plMax < 11) ? alpha1 : alpha2;

                dst[i] = new double[Pl[i].Length];

                for (int j = 0; j < orders2.Length; j++)
                {
                    dst[i][j] += thisAlpha * poly[i][j];
                    dst[i][orders2[j]] += (1 - thisAlpha) / (j + 1f);
                }
            }

            return dst;
        }

        public static double[][] PolyMerge2(
            double[][] T3, double[][] T3Hyb1, double[][] Pl, double[][] PlHyb1,
            ushort[] personal0,
            double alpha, double t3Alpha, double plAlpha, double persAlpha)
        {
            double[][] dst = new double[Pl.Length][];

            for (int i = 0; i < Pl.Length; i++)
            {
                int[] orders1 = T3[i].MyOrderDescending();
                int[] orders2 = T3Hyb1[i].MyOrderDescending();
                int[] orders3 = Pl[i].MyOrderDescending();
                int[] orders4 = PlHyb1[i].MyOrderDescending();

                double baseAlpha = ((personal0[i] < 1200)
                    //&& ((personal0[i] < 750) || (personal0[i] > 820))
                    && (personal0[i] >= 30)) ? persAlpha : alpha;

                double alphaTop1Val = Math.Max(0, Pl[i][orders3[0]] - 1);
                double thisAlpha = baseAlpha + (1 - baseAlpha) / (1 + alphaTop1Val);

                double t3Hyb1Top1Val = Math.Max(0, T3Hyb1[i][orders2[0]] - 1);
                double t3Hyb1Alpha = (t3Alpha + (1 - t3Alpha) / (1 + t3Hyb1Top1Val));

                double plHyb1Top1Val = Math.Log(Math.Max(1, PlHyb1[i][orders4[0]]));
                double plHyb1Alpha = (plAlpha + (1 - plAlpha) / (1 + plHyb1Top1Val));

                dst[i] = new double[Pl[i].Length];

                for (int j = 0; j < orders3.Length; j++)
                {
                    dst[i][orders1[j]] += thisAlpha * t3Hyb1Alpha / (j + 1f);

                    if (T3Hyb1[i][orders2[j]] > 1)
                    {
                        dst[i][orders2[j]] += thisAlpha * (1 - t3Hyb1Alpha) / (j + 1f);
                    }

                    if (Pl[i][orders3[j]] >= 0)
                    {
                        dst[i][orders3[j]] += (1 - thisAlpha) * plHyb1Alpha / (j + 1f);

                        if (PlHyb1[i][orders4[j]] >= 0)
                        {
                            dst[i][orders4[j]] += (1 - thisAlpha) * (1 - plHyb1Alpha) / (j + 1f);
                        }
                    }
                }

                for (int j = 0; j < orders3.Length; j++)
                {
                    dst[i][j] = dst[i][j] * (1 + (Pl[i][j] - (int)(Pl[i][j] + 0.5)));
                }
            }

            return dst;
        }

        public static double[][] PolyMerge3(
            double[][] HighT3, double[][] T3Hyb1, double[][] Pl, double[][] PlHyb1,
            ushort[] personal0,
            double alpha, double t3Alpha, double plAlpha, double persAlpha)
        {
            double[][] dst = new double[Pl.Length][];

            for (int i = 0; i < Pl.Length; i++)
            {
                int[] highOrders1 = HighT3[i].MyOrderDescending();
                int[] orders2 = T3Hyb1[i].MyOrderDescending();
                int[] orders3 = Pl[i].MyOrderDescending();
                int[] orders4 = PlHyb1[i].MyOrderDescending();

                double baseAlpha = ((personal0[i] < 1200)
                    //&& ((personal0[i] < 750) || (personal0[i] > 820))
                    && (personal0[i] >= 30)) ? persAlpha : alpha;

                double alphaTop1Val = Math.Max(0, Pl[i][orders3[0]] - 1);
                double thisAlpha = baseAlpha + (1 - baseAlpha) / (1 + 0.5 * alphaTop1Val);

                double t3Hyb1Top1Val = Math.Max(0, T3Hyb1[i][orders2[0]] - 1);
                double t3Hyb1Alpha = (t3Alpha + (1 - t3Alpha) / (1 + t3Hyb1Top1Val));

                double plHyb1Top1Val = Math.Log(Math.Max(1, PlHyb1[i][orders4[0]]));
                double plHyb1Alpha = (plAlpha + (1 - plAlpha) / (1 + plHyb1Top1Val));

                dst[i] = new double[Pl[i].Length];

                for (int j = 0; j < orders3.Length; j++)
                {
                    dst[i][highOrders1[j]] += thisAlpha * t3Hyb1Alpha / (j + 1f);

                    if (T3Hyb1[i][orders2[j]] > 1)
                    {
                        dst[i][orders2[j]] += thisAlpha * (1 - t3Hyb1Alpha) / (j + 1f);
                    }

                    if (Pl[i][orders3[j]] >= 0)
                    {
                        dst[i][orders3[j]] += (1 - thisAlpha) * plHyb1Alpha / (j + 1f);

                        if (PlHyb1[i][orders4[j]] >= 0)
                        {
                            dst[i][orders4[j]] += (1 - thisAlpha) * (1 - plHyb1Alpha) / (j + 1f);
                        }
                    }
                }

                for (int j = 0; j < orders3.Length; j++)
                {
                    dst[i][j] = dst[i][j] * (1 + (Pl[i][j] - (int)(Pl[i][j] + 0.5)));
                }
            }

            return dst;
        }

        public static double[][] PolyMerge4(double[][] T3, double[][] T3Hyb1, double t3Alpha)
        {
            double[][] dst = new double[T3.Length][];

            for (int i = 0; i < T3.Length; i++)
            {
                int[] orders1 = T3[i].MyOrderDescending();
                int[] orders2 = T3Hyb1[i].MyOrderDescending();

                double t3Hyb1Top1Val = Math.Max(0, T3Hyb1[i][orders2[0]] - 1);
                double t3Hyb1Alpha = (t3Alpha + (1 - t3Alpha) / (1 + t3Hyb1Top1Val));


                dst[i] = new double[T3[i].Length];

                for (int j = 0; j < orders1.Length; j++)
                {
                    dst[i][orders1[j]] += t3Hyb1Alpha / (j + 1f);

                    if (T3Hyb1[i][orders2[j]] > 1)
                    {
                        dst[i][orders2[j]] += (1 - t3Hyb1Alpha) / (j + 1f);
                    }
                }
            }

            return dst;
        }
    }
}
