﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Threading.Tasks;
using GenEngAttr.Data;

namespace GenEngAttr.Utils
{
    // non project specific utility methods
    public static class Util
    {
        public static string FileDateTimeFormat = "yyyyMMddHHmmss";

        public static void Print(string format, params object[] args)
        {
            string str = string.Format(format, args);

            System.Console.WriteLine(str);
        }

        public static ParallelLoopResult MyParallelFor(int maxThreads, int fromInclusive, int toExclusive, Action<int> body)
        {
            System.Threading.SemaphoreSlim maxThread4 = new System.Threading.SemaphoreSlim(maxThreads);

            return Parallel.For(fromInclusive, toExclusive, (i) =>
            {
                maxThread4.Wait();

                try
                {
                    body(i);
                }
                finally
                {
                    maxThread4.Release();
                }
            });
        }

        public static void CreateDirIfNotExists(string dirPath) // create all dirs if not exist
        {
            if (!System.IO.Directory.Exists(dirPath))
            {
                System.IO.Directory.CreateDirectory(dirPath);
            }
        }

        public static string MyToSecString(this DateTime dateTime) // todo remove
        {
            return dateTime.ToString("yyyy-MM-dd HH:mm:ss.fff");
        }

        public static Tuple<T1, T2> MyTuple<T1, T2>(T1 item1, T2 item2)
        {
            return new Tuple<T1, T2>(item1, item2);
        }

        public static Tuple<T1, T2, T3> MyTuple<T1, T2, T3>(T1 item1, T2 item2, T3 item3)
        {
            return new Tuple<T1, T2, T3>(item1, item2, item3);
        }

        public static Tuple<T1, T2, T3, T4> MyTuple<T1, T2, T3, T4>(T1 item1, T2 item2, T3 item3, T4 item4)
        {
            return new Tuple<T1, T2, T3, T4>(item1, item2, item3, item4);
        }

        public static T[] InitArray<T>(T initValue, int n)
        {
            T[] array = new T[n];

            for (int i = 0; i < n; i++)
            {
                array[i] = initValue;
            }

            return array;
        }

        public static T[] InitArrayStep<T>(T initValue, T stepValue, int n)
        {
            T[] array = new T[n];

            for (int i = 0; i < n; i++)
            {
                array[i] = initValue + i * (dynamic)stepValue;
            }

            return array;
        }

        public static void AddInPlace(double[][] dstsrc1, double[][] src2)
        {
            for (int i = 0; i < dstsrc1.Length; i++)
            {
                for (int j = 0; j < dstsrc1[i].Length; j++)
                {
                    dstsrc1[i][j] += src2[i][j];
                }
            }
        }

        public static int[] MyOrder<T>(this IEnumerable<T> source)
        {
            T[] items = source.ToArray();
            int[] orders = new int[items.Length];

            for (int i = 0; i < orders.Length; i++)
            {
                orders[i] = i;
            }

            Array.Sort(items, orders);

            return orders;
        }

        public static int[] MyOrderDescending<T>(this IEnumerable<T> source)
        {
            T[] items = source.ToArray();
            int[] orders = new int[items.Length];

            for (int i = 0; i < orders.Length; i++)
            {
                orders[i] = i;
            }

            Array.Sort(items, orders);

            return orders.Reverse().ToArray();
        }

        public static int[] InitDevSet(int len, int seed, int iPart, double share)
        {
            Random random = new Random(seed);
            double minRange = share * iPart;
            double maxRange = share * (iPart + 1);
            List<int> iList = new List<int>();

            for (int i = 0; i < len; i++)
            {
                double val = random.NextDouble();

                if ((val >= minRange) && (val < maxRange))
                {
                    iList.Add(i);
                }
            }

            return iList.ToArray();
        }

        public static bool[] GetDisables(int length, int[] devSet)
        {
            bool[] disables = new bool[length];

            for (int o = 0; o < devSet.Length; o++)
            {
                disables[devSet[o]] = true;
            }

            return disables;
        }

        public static T[][] VectorToMatrix<T>(T[] src, int n)
        {
            int nLength = src.Length / n;
            T[][] dst = new T[nLength][];

            for (int i = 0; i < nLength; i++)
            {
                dst[i] = new T[n];

                for (int j = 0; j < n; j++)
                {
                    dst[i][j] = src[i * n + j];
                }
            }

            return dst;
        }

        public static T[] MatrixToVector<T>(T[][] src)
        {
            int n = src[0].Length;
            T[] dst = new T[src.Length * n];

            for (int i = 0; i < src.Length; i++)
            {
                for (int j = 0; j < n; j++)
                {
                    dst[i * n + j] = src[i][j];
                }
            }

            return dst;
        }

        // https://en.wikipedia.org/wiki/Hamming_weight
        const ulong m1 = 0x5555555555555555; //binary: 0101...
        const ulong m2 = 0x3333333333333333; //binary: 00110011..
        const ulong m4 = 0x0f0f0f0f0f0f0f0f; //binary:  4 zeros,  4 ones ...

        public static int PopCount(ulong x)
        {
            x -= (x >> 1) & m1;             //put count of each 2 bits into those 2 bits
            x = (x & m2) + ((x >> 2) & m2); //put count of each 4 bits into those 4 bits 
            x = (x + (x >> 4)) & m4;        //put count of each 8 bits into those 8 bits 
            x += x >> 8;  //put count of each 16 bits into their lowest 8 bits
            x += x >> 16;  //put count of each 32 bits into their lowest 8 bits
            x += x >> 32;  //put count of each 64 bits into their lowest 8 bits

            return (int)(x & 0x7f);
        }

        public static void WriteSimple(string path, float[] points)
        {
            int rawSize = System.Runtime.InteropServices.Marshal.SizeOf(typeof(float));
            byte[] rawData = new byte[rawSize * points.Length];
            int offset = 0;

            foreach (float point in points)
            {
                byte[] pointBytes = BitConverter.GetBytes(point);

                rawData[offset + 0] = pointBytes[0];
                rawData[offset + 1] = pointBytes[1];
                rawData[offset + 2] = pointBytes[2];
                rawData[offset + 3] = pointBytes[3];

                offset += rawSize;
            }

            File.WriteAllBytes(path, rawData);
        }

        public static void WriteSimple(string path, double[] points)
        {
            int rawSize = System.Runtime.InteropServices.Marshal.SizeOf(typeof(double));
            byte[] rawData = new byte[rawSize * points.Count()];
            int offset = 0;

            foreach (double point in points)
            {
                byte[] pointBytes = BitConverter.GetBytes(point);

                rawData[offset + 0] = pointBytes[0];
                rawData[offset + 1] = pointBytes[1];
                rawData[offset + 2] = pointBytes[2];
                rawData[offset + 3] = pointBytes[3];
                rawData[offset + 4] = pointBytes[4];
                rawData[offset + 5] = pointBytes[5];
                rawData[offset + 6] = pointBytes[6];
                rawData[offset + 7] = pointBytes[7];

                offset += rawSize;
            }

            File.WriteAllBytes(path, rawData);
        }

        public static void WriteSimple(string path, ushort[] points)
        {
            int rawSize = System.Runtime.InteropServices.Marshal.SizeOf(typeof(ushort));
            byte[] rawData = new byte[rawSize * points.Count()];
            int offset = 0;

            foreach (ushort point in points)
            {
                byte[] pointBytes = BitConverter.GetBytes(point);

                rawData[offset + 0] = pointBytes[0];
                rawData[offset + 1] = pointBytes[1];

                offset += rawSize;
            }

            File.WriteAllBytes(path, rawData);
        }

        public static void WriteSimple(string path, int[] points)
        {
            int rawSize = System.Runtime.InteropServices.Marshal.SizeOf(typeof(int));
            byte[] rawData = new byte[rawSize * points.Count()];
            int offset = 0;

            foreach (int point in points)
            {
                byte[] pointBytes = BitConverter.GetBytes(point);

                rawData[offset + 0] = pointBytes[0];
                rawData[offset + 1] = pointBytes[1];
                rawData[offset + 2] = pointBytes[2];
                rawData[offset + 3] = pointBytes[3];

                offset += rawSize;
            }

            File.WriteAllBytes(path, rawData);
        }

        public static void WriteSimple(string path, ulong[] points)
        {
            int rawSize = System.Runtime.InteropServices.Marshal.SizeOf(typeof(ulong));
            byte[] rawData = new byte[rawSize * points.Count()];
            int offset = 0;

            foreach (ulong point in points)
            {
                byte[] pointBytes = BitConverter.GetBytes(point);

                rawData[offset + 0] = pointBytes[0];
                rawData[offset + 1] = pointBytes[1];
                rawData[offset + 2] = pointBytes[2];
                rawData[offset + 3] = pointBytes[3];
                rawData[offset + 4] = pointBytes[4];
                rawData[offset + 5] = pointBytes[5];
                rawData[offset + 6] = pointBytes[6];
                rawData[offset + 7] = pointBytes[7];

                offset += rawSize;
            }

            File.WriteAllBytes(path, rawData);
        }

        public static float[] ReadSimpleFloat(string path)
        {
            byte[] rawData = File.ReadAllBytes(path);
            int rawSize = System.Runtime.InteropServices.Marshal.SizeOf(typeof(float));
            int count = rawData.Length / rawSize;
            float[] points = new float[count];
            int offset = 0;

            for (int i = 0; i < count; i++)
            {
                float point = BitConverter.ToSingle(rawData, offset);

                points[i] = point;

                offset += rawSize;
            }

            return points;
        }

        public static double[] ReadSimpleDouble(string path)
        {
            byte[] rawData = File.ReadAllBytes(path);
            int rawSize = System.Runtime.InteropServices.Marshal.SizeOf(typeof(double));
            int count = rawData.Length / rawSize;
            double[] points = new double[count];
            int offset = 0;

            for (int i = 0; i < count; i++)
            {
                double point = BitConverter.ToDouble(rawData, offset);

                points[i] = point;

                offset += rawSize;
            }

            return points;
        }

        public static int[] ReadSimpleInt(string path)
        {
            byte[] rawData = File.ReadAllBytes(path);
            int rawSize = System.Runtime.InteropServices.Marshal.SizeOf(typeof(int));
            int count = rawData.Length / rawSize;
            int[] points = new int[count];
            int offset = 0;

            for (int i = 0; i < count; i++)
            {
                int point = BitConverter.ToInt32(rawData, offset);

                points[i] = point;

                offset += rawSize;
            }

            return points;
        }

        public static ushort[] ReadSimpleUShort(string path)
        {
            byte[] rawData = File.ReadAllBytes(path);
            int rawSize = System.Runtime.InteropServices.Marshal.SizeOf(typeof(ushort));
            int count = rawData.Length / rawSize;
            ushort[] points = new ushort[count];
            int offset = 0;

            for (int i = 0; i < count; i++)
            {
                ushort point = BitConverter.ToUInt16(rawData, offset);

                points[i] = point;

                offset += rawSize;
            }

            return points;
        }

        public static ulong[] ReadSimpleULong(string path)
        {
            byte[] rawData = File.ReadAllBytes(path);
            int rawSize = System.Runtime.InteropServices.Marshal.SizeOf(typeof(ulong));
            int count = rawData.Length / rawSize;
            ulong[] points = new ulong[count];
            int offset = 0;

            for (int i = 0; i < count; i++)
            {
                ulong point = BitConverter.ToUInt64(rawData, offset);

                points[i] = point;

                offset += rawSize;
            }

            return points;
        }
    }
}
