import numpy as np
from random import choice


class Dataset:
    def __init__(self, data, max_len):
        self.data = data.reset_index(drop=True).copy()
        self.num_fea = [x for x in data if (x != 'target') and ('sequence' not in x) and (x != 'rev_seq')]
        self.max_len = max_len
        assert len(self.num_fea) == 39

    def __len__(self):
        return self.data.shape[0]

    def __getitem__(self, item):
        row = self.data.iloc[item]
        seq = np.array(row['sequence_2gramo'], dtype=np.int64)
        rev_seq = np.array(row['rev_seq'], dtype=np.int64)

        if len(seq) < self.max_len:
            cali_seq = np.hstack([seq, np.zeros(self.max_len-len(seq), dtype=np.int64)])
        else:
            start_index = choice(range(len(seq)-self.max_len+1))
            cali_seq = seq[start_index:(start_index+self.max_len)]

        if len(rev_seq) < self.max_len:
            cali_seq_rev = np.hstack([rev_seq, np.zeros(self.max_len-len(rev_seq), dtype=np.int64)])
        else:
            start_index = choice(range(len(rev_seq)-self.max_len+1))
            cali_seq_rev = rev_seq[start_index:(start_index+self.max_len)]

        result = {'sequence': cali_seq,
                  'rev_seq': cali_seq_rev,
                  'num_fea': row[self.num_fea].values.astype(np.float32)}
        if 'target' in row:
            result['target'] = row['target'].astype(np.int64)

        return result
