import pandas as pd
from tqdm.auto import tqdm
from tensorflow.keras.preprocessing.text import Tokenizer
from time import time
import pickle5 as pickle
from Bio.Seq import Seq

n = 8

train_label = pd.read_csv("train_labels.csv", index_col='sequence_id')
lab_ids = pd.DataFrame(train_label.idxmax(axis=1), columns=['lab_id']).reset_index()
label_map = dict(zip(train_label.columns.tolist(), range(1314)))
lab_ids['target'] = lab_ids['lab_id'].apply(lambda x: label_map[x])

train_fea = pd.read_csv("train_values.csv")
test_fea = pd.read_csv("test_values.csv")
train_fea['target'] = lab_ids['target']


ngram_train = []
ngram_train_rev = []
for i in tqdm(range(train_fea.shape[0])):
    seq1 = train_fea['sequence'].iloc[i]
    rev_seq = str(Seq(seq1).reverse_complement())
    ngram_train.append([seq1[i:(i + n)] for i in range(0, len(seq1)-n+1)])
    ngram_train_rev.append([rev_seq[i:(i + n)] for i in range(0, len(rev_seq)-n+1)])

ngram_test = []
ngram_test_rev = []
for i in tqdm(range(test_fea.shape[0])):
    seq1 = test_fea['sequence'].iloc[i]
    rev_seq = str(Seq(seq1).reverse_complement())
    ngram_test.append([seq1[i:(i + n)] for i in range(0, len(seq1)-n+1)])
    ngram_test_rev.append([rev_seq[i:(i + n)] for i in range(0, len(rev_seq) - n + 1)])

start_time = time()
tokenizer = Tokenizer(num_words=100000, lower=False, oov_token=None)
tokenizer.fit_on_texts(ngram_train + ngram_test + ngram_train_rev + ngram_test_rev)
print(f"elapsed time: {time() - start_time}")
train_fea['sequence_2gramo'] = tokenizer.texts_to_sequences(ngram_train)
print(f"elapsed time: {time() - start_time}")
test_fea['sequence_2gramo'] = tokenizer.texts_to_sequences(ngram_test)
print(f"elapsed time: {time() - start_time}")
print(f"elapsed time: {time() - start_time}")
train_fea['rev_seq'] = tokenizer.texts_to_sequences(ngram_train_rev)
print(f"elapsed time: {time() - start_time}")
test_fea['rev_seq'] = tokenizer.texts_to_sequences(ngram_test_rev)
print(f"elapsed time: {time() - start_time}")

del train_fea['sequence'], test_fea['sequence']
pickle.dump(train_fea, open("train_fea.pk", "wb"), protocol=5)
pickle.dump(test_fea, open("test_fea.pk", "wb"), protocol=5)

print("correctly transform the verification tokens")
test_fea = pd.read_csv("verification_values.csv")
ngram_test = []
ngram_test_rev = []
for i in tqdm(range(test_fea.shape[0])):
    seq1 = test_fea['sequence'].iloc[i]
    rev_seq = str(Seq(seq1).reverse_complement())
    ngram_test.append([seq1[i:(i + n)] for i in range(0, len(seq1)-n+1)])
    ngram_test_rev.append([rev_seq[i:(i + n)] for i in range(0, len(rev_seq) - n + 1)])
test_fea['sequence_2gramo'] = tokenizer.texts_to_sequences(ngram_test)
test_fea['rev_seq'] = tokenizer.texts_to_sequences(ngram_test_rev)
del test_fea['sequence']
pickle.dump(test_fea, open("test_fea.pk", "wb"), protocol=5)