import pandas as pd
import numpy as np
import os
from sklearn.model_selection import StratifiedKFold
from nntrainer import NNTrainer
from nnmodel import JointModel
import pickle5 as pickle
from sys import argv

max_len = int(argv[1])
load_pretrained = argv[2] == 'true'
print(f"loading pretrained models: {load_pretrained}")
NFOLDS = 5
NSEED = 1 if load_pretrained else 5

if load_pretrained:
    prefix = f'8gram{int(max_len/1000)}'
else:
    prefix = f'8gram{max_len}'
    print("creating folder")
    os.mkdir(f"{prefix}")
    for i in range(5):
        os.mkdir(f"{prefix}/fold{i}")
        for s in range(NSEED):
            os.mkdir(f"{prefix}/fold{i}/seed{s}")


def top10_acc(y_pred, y):
    top10 = np.argpartition(y_pred, -10, axis=1)[:, -10:]
    mask = top10 == y.reshape((y.size, 1))
    return mask.any(axis=1).mean()


train_label = pd.read_csv("train_labels.csv", index_col='sequence_id')
train_fea = pickle.load(open("train_fea.pk", "rb"))
test_fea = pickle.load(open("test_fea.pk", "rb"))

cv_train = np.zeros((train_fea.shape[0], 1314))
cv_pred = np.zeros((test_fea.shape[0], 1314))
fold_scores = []

for seed in range(NSEED):
    kfold = StratifiedKFold(n_splits=NFOLDS, shuffle=True, random_state=218+seed)
    kf = kfold.split(train_fea, train_fea['target'])
    for i, (train_fold, validate) in enumerate(kf):
        print("##############################")
        print(f"training fold {i} seed {seed}......")
        print("##############################")

        model = NNTrainer(model_cls=JointModel(), model_path=f"{prefix}/fold{i}/seed{seed}", max_len=max_len)
        if not load_pretrained:
            model.fit(train_fea.iloc[train_fold], train_fea.iloc[validate], batch_size=64, num_epochs=30, early_stop=5)
        cv_train[validate] += model.predict(train_fea.iloc[validate], batch_size=64)/NSEED
        c_fscore = top10_acc(cv_train[validate], train_fea['target'].values[validate])
        print(seed, c_fscore)
        cv_pred += model.predict(test_fea, batch_size=64)/NSEED

cv_pred /= NFOLDS

cv_score = top10_acc(cv_train, train_fea['target'].values)
print(f"cv score: {cv_score}")

submission = pd.DataFrame(data=cv_pred, columns=train_label.columns.tolist())
submission = pd.concat([test_fea[['sequence_id']], submission], axis=1)
submission.to_csv(f"{prefix}.csv", index=False, float_format='%.15f')
