[<img src='https://s3.amazonaws.com/drivendata-public-assets/logo-white-blue.png' width='600'>](https://www.drivendata.org/)
<br><br>

![Banner Image](https://s3.amazonaws.com/drivendata-public-assets/al-green-homepage.jpg)


# Genetic Engineering Attribution Challenge

## Goal of the Competition

Genetic engineering is the process of modifying an organism's DNA to alter its function. Its use has become increasingly widespread, accessible, and powerful. The benefits of this progress is undeniable, but the tools to ensure transparency and accountability are lagging. Currently, there is no easy way of tracing genetically engineered DNA back to its lab-of-origin. This task is known as attribution, and it's a pivotal part of ensuring that genetic engineering progresses responsibly.

When manipulating DNA, a designer has many decisions she must make, such as promoter choice and cloning method. These choices leave clues in the genetic material, and together, compose a "genetic fingerprint" that can be traced back to the designer.

The goal of the Genetic Engineering Attribution Challenge was to develop tools that help human decision makers identify the lab-of-origin from genetically engineered DNA. This competition was composed of two tracks. In the Prediction Track, participants competed to attribute DNA samples to its lab-of-origin with the highest possible top-10 accuracy. In the Innovation Track, competitors that beat the BLAST benchmark were invited to submit reports demonstrating how their lab-of-origin prediction models excel in domains beyond raw accuracy.


## What's in this Repository

This repository contains code from winning competitors in the [Genetic Engineering Attribution Challenge](https://www.drivendata.org/competitions/63/genetic-engineering-attribution/) DrivenData challenge.

**Winning code for other DrivenData competitions is available in the [competition-winners repository](https://github.com/drivendataorg/competition-winners).**

## Winning Submissions

Prediction Track Place | Innovation Track Place | Team or User | Private Score | Public Score | Summary of Model
---   | ---| ---          | ---          | ---           | ---
1     |1 | [sorrge](https://www.drivendata.org/users/sorrge/)    | 0.949 | 0.950 | I made two kinds of models. The first one is based on k-mer counts, because I know from experience that simple k-mer approaches are powerful for DNA sequence matching and comparison. The second is a convolutional neural network with capacity as large as I could make. The two kinds of models complement each other's strengths well: k-mers are good at precise sequence matching, and work well with low number of samples. CNNs are good at fuzzy motif finding and integrating weak evidence from multiple sites in the sequence. Among the 7 CNNs in the ensemble, 5 are trained directly to predict the lab index, and 2 use self-supervised pre-training of the main CNN layer. The self-supervised target is to predict whether two sequences come from the same lab, using NT-Xent loss.
2     | 4 | fit_dna: [ZFTurbo](https://www.drivendata.org/users/ZFTurbo/), [bragin](https://www.drivendata.org/users/bragin/)   | 0.944 | 0.956 | Our main idea was to find popular subsequences which describe a lab-of-origin. We tested two approaches: 1) Use DNN for automatic sequence analysis: We built a DNN (deep neural network) architecture which automatically extracts subsequences and classifies them. 2) Based on BLAST output: If BLAST finds a similar subsequence for any pair of DNA sequences, we cut it and add it to a set of interesting subsequences. Then we build a matrix of test+train size by number of sequences, about 80,000 by 10^7. This is too big for training. After filtering only popular subsequences and applying Principal Component Analysis (PCA), we extract 1,024 main components and successfully use them to build the lab-of-origin classifier. There are issues in this approach. BLAST doesn’t find all matches in a sequence pair, it reports only the longest one. Furthermore, analysis of 10^7 subsequences requires a lot of hardware resources so we had to reduce it significantly, and thus lost a lot of information.
3     |N/A | [eakmail](https://www.drivendata.org/users/eakmail/)    | 0.934 | 0.947 | I used various n-gram (k-mer in biology) kernels, naive Bayes, soft masks, and rank model merging. I mostly go for fast solutions. I got to the top 7 on my first day with some one-hour n-gram solution. Later I tweaked the kernel to be +4% better and reduced the time down to 4 minutes.
4     |N/A| What a Team: [BestGCE](https://www.drivendata.org/users/BestGCE/), [KarmaML](https://www.drivendata.org/users/KarmaML/) | 0.930 | 0.942 | Yao: I used 100% CNN models since I have been learning deep learning recently. I tried different k-mers but found that 8-mers worked the best. I modeled my solution after this paper: Convolutional Neural Networks for Sentence Classification. Grigorii: I‘m using a simple MLP model, trained on k-mer counts and phenotypic features. It shows high accuracy scores, is quick to train, and doesn’t require a high-end PC. For feature generation, I counted all 6-mers after extending sequences with its reverse complement. Adding the reverse complement improved quality a lot, and the number of features was reduced after dropping duplicate columns. I also scored using average accuracy per class, which helps to score rare classes.
N/A     |2| Amalgam: [adriano_xnv](https://www.drivendata.org/users/adriano_xnv/), [fernandocamargo](https://www.drivendata.org/users/fernandocamargo/), [IgorMuniz](https://www.drivendata.org/users/IgorMuniz/)  | 0.917 | 0.938 | Instead of doing a common multi-class classification with a softmax layer, we used Triplet Networks in order to learn embeddings for the labs and learn an embedding extractor for DNA sequences. In this way, we can use distance metrics to rank the labs-of-origin while having the embeddings as a useful sub-product. The dataset gives us the DNA sequences, some extra inputs, and the lab-of-origin for each of these sequences. We use the DNA sequences and extra inputs to extract an embedding for each sequence. We also have an embedding layer for the labs. We use the lab we get from the dataset as our positive example. Then we do hard negative mining to generate the negative examples. In this case, the negative example is a lab that is the closest to the DNA sequence but is not the lab-of-origin.
N/A     |3 | [neurogenesis](https://www.drivendata.org/users/neurogenesis/) | 0.788 | 0.828 | I moved the problem to another dimension by transforming 1D sequences into a 2D image and exploited new relations (statistics, feature extraction, etc.).

Additional solution details can be found in the `reports` folder inside the directory for each submission.
