# coding: utf-8
__author__ = 'ZFTurbo: https://www.drivendata.org/users/ZFTurbo/'


from utils.a00_common_functions import *


def create_matrix_v1(run_type, type):
    # from scipy.sparse import csr_matrix, save_npz
    headers = ['query_id', 'subject_id', 'identity', 'alignment length', 'mismatches', 'gap opens', 'q.start',
               'q.end', 's.start', 's.end', 'evalue', 'bit score']

    train = pd.read_csv(INPUT_PATH + 'train_values.csv')
    test = pd.read_csv(INPUT_PATH + 'test_values.csv')
    holdout = pd.read_csv(INPUT_PATH + 'holdout_values.csv')

    train_ids = train['sequence_id'] + '_train'
    test_ids = test['sequence_id'] + '_test'
    holdout_ids = holdout['sequence_id'] + '_holdout'
    ids = list(train_ids.values) + list(test_ids.values)

    ids_map_train_test = dict()
    for i, id in enumerate(ids):
        ids_map_train_test[id] = i
    print('Feature vector size: {}'.format(len(ids_map_train_test)))

    ids_map_holdout = dict()
    for i, id in enumerate(list(holdout_ids.values)):
        ids_map_holdout[id] = i
    print('Holdout size: {}'.format(len(ids_map_holdout)))

    files = glob.glob(CACHE_PATH + '/blast_res_holdout_vs_full/*.csv')
    if type == 'alen' or type == 'mismatches' or type == 'gap':
        matrix = np.zeros((len(holdout_ids), len(ids)), dtype=np.uint32)
    else:
        matrix = np.zeros((len(holdout_ids), len(ids)), dtype=np.float32)

    pairs = dict()
    for f in files:
        print('Go for: {}'.format(f))
        in1 = open(f, 'r')
        while 1:
            line = in1.readline()
            if not line:
                break
            arr = line.split('\t')
            if type == 'alen':
                qid, sid, var = arr[0], arr[1], int(arr[3])
            elif type == 'bitscore':
                qid, sid, var = arr[0], arr[1], float(arr[-1].strip())
            elif type == 'identity':
                qid, sid, var = arr[0], arr[1], float(arr[2])
            elif type == 'mismatches':
                qid, sid, var = arr[0], arr[1], int(arr[4]) + 1
            elif type == 'gap':
                qid, sid, var = arr[0], arr[1], int(arr[5]) + 1
            elif type == 'evalue':
                qid, sid, var = arr[0], arr[1], float(arr[-2])
            else:
                print('Unknown type: {}'.format(type))
                exit()
            # print(qid, sid, var)
            sh0 = ids_map_holdout[qid]
            sh1 = ids_map_train_test[sid]
            matrix[sh0, sh1] = var
        in1.close()

    print('Matrix sum: {}'.format(matrix.sum()))
    save_in_file((holdout_ids, matrix), CACHE_PATH + 'pca_{}_{}.pklz'.format(type, run_type))
    # matrix = csr_matrix(matrix)
    # save_npz(CACHE_PATH + 'matrix_blast_alength_{}_sparse.npz'.format(use_large), matrix)


def create_log_matrix_alen(run_type):
    path = CACHE_PATH + 'pca_alen_{}.pklz'.format(run_type)
    ids, matrix = load_from_file(path)
    matrix = np.log1p(matrix)
    save_in_file((ids, matrix), os.path.dirname(path) + '/pca_alenlog_{}.pklz'.format(run_type))


def create_log_matrix_bitscore(run_type):
    path = CACHE_PATH + 'pca_bitscore_{}.pklz'.format(run_type)
    ids, matrix = load_from_file(path)
    matrix = np.log1p(matrix)
    save_in_file((ids, matrix), os.path.dirname(path) + '/pca_bitscorelog_{}.pklz'.format(run_type))


if __name__ == '__main__':
    run_type = 'holdout'
    create_matrix_v1(run_type, 'alen')
    create_matrix_v1(run_type, 'bitscore')
    create_matrix_v1(run_type, 'identity')
    create_matrix_v1(run_type, 'mismatches')
    create_matrix_v1(run_type, 'gap')
    create_log_matrix_alen(run_type)
    create_log_matrix_bitscore(run_type)
