# coding: utf-8
__author__ = 'ZFTurbo: https://www.drivendata.org/users/ZFTurbo/'


from utils.a00_common_functions import *


# NUMBER_TO_FIND = 999
NUM_SPLITS = 8
# PATH_TO_BLAST = "D://blast/bin//blastn.exe"
# PATH_TO_MAKEDB = "D://blast/bin//makeblastdb.exe"
PATH_TO_BLAST = "blastn"
PATH_TO_MAKEDB = "makeblastdb"

if len(sys.argv) > 1:
    PATH_TO_BLAST = sys.argv[1]
    PATH_TO_MAKEDB = sys.argv[2]

print('Blast binary path: {}'.format(PATH_TO_BLAST))
print('MakeDB binary path: {}'.format(PATH_TO_MAKEDB))


def create_task_for_train(number_to_find):
    task_type = 'full_vs_full_{}'.format(number_to_find)
    num_splits = NUM_SPLITS
    out_folder = CACHE_PATH + 'blast_task_{}/'.format(task_type)
    if not os.path.isdir(out_folder):
        os.mkdir(out_folder)
    print('Out folder: {}'.format(out_folder))

    train = pd.read_csv(INPUT_PATH + 'train_values.csv')
    train['type'] = 'train'
    test = pd.read_csv(INPUT_PATH + 'test_values.csv')
    test['type'] = 'test'
    train['sequence_id'] = train['sequence_id'] + '_' + train['type']
    test['sequence_id'] = test['sequence_id'] + '_' + test['type']
    full_data = pd.concat([train[['sequence_id', 'sequence']], test[['sequence_id', 'sequence']]], axis=0)

    print(train.shape, test.shape, full_data.shape)

    # make the train sequences into a temporary fasta file with numerical index
    with open(out_folder + 'train_fold_{}.fasta'.format(task_type), 'w+') as f:
        part = full_data
        for i in range(len(part)):
            f.write('>{}\n'.format(part["sequence_id"].values[i]))
            f.write('{}\n'.format(part["sequence"].values[i]))

    part = full_data
    for j in range(0, num_splits):
        start = j * (len(part) // num_splits)
        if j == num_splits - 1:
            end = len(part)
        else:
            end = (j + 1) * (len(part) // num_splits)
        subpart = part[start:end]
        with open(out_folder + 'query_fold_{}_part_{}.fasta'.format(task_type, j), 'w+') as f:
            for i in range(len(subpart)):
                f.write('>{}\n'.format(subpart["sequence_id"].values[i]))
                f.write('{}\n'.format(subpart["sequence"].values[i]))

    out_file1 = out_folder + 'step1_create_db.bat'
    out = open(out_file1, 'w')
    out.write(PATH_TO_MAKEDB + ' -in train_fold_{}.fasta -parse_seqids -title "train_set_{}" -dbtype nucl'.format(task_type, task_type))
    out.close()
    print('File created: {}'.format(out_file1))

    for j in range(0, num_splits):
        in_path = out_folder.replace('\\', '/') + 'query_fold_{}_part_{}.fasta'.format(task_type, j)
        if not os.path.isdir(CACHE_PATH + 'blast_res_full_vs_full_{}'.format(number_to_find)):
            os.mkdir(CACHE_PATH + 'blast_res_full_vs_full_{}'.format(number_to_find))

        out_path = CACHE_PATH.replace('\\', '/') + 'blast_res_full_vs_full_{}/result_fold_{}_part_{}.csv'.format(number_to_find, task_type, j)
        out_file2 = out_folder + 'step2_run_{}.bat'.format(j)
        out = open(out_file2, 'w')
        run_str = PATH_TO_BLAST + ' -db train_fold_{}.fasta -query {} -out {} -outfmt 6 -max_target_seqs {} -evalue 10 -max_hsps=1 -num_threads 4'.format(task_type, in_path, out_path, number_to_find)
        out.write(run_str)
        out.close()
        print('File created: {}'.format(out_file2))


if __name__ == '__main__':
    create_task_for_train(999)
    create_task_for_train(9999999)
